import itertools as it
import random

import pytest

from hstrat import hstrat


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95, 0.99],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_rank_of_earliest_detectable_mrca_among1(
    confidence_level,
    differentia_bit_width,
    wrap,
):
    expected_thresh = (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_bit_width,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1 - confidence_level,
        )
        - 1
    )

    for r1, r2, r3 in it.product(*[range(1, expected_thresh)] * 3):
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_bit_width,
            stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
        )
        c2 = c1.Clone()
        c3 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_bit_width,
            stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
        )
        for __ in range(r1 - 1):
            c1.DepositStratum()
        for __ in range(r2 - 1):
            c2.DepositStratum()
        for __ in range(r3 - 1):
            c3.DepositStratum()

        for x1, x2 in it.combinations([c1, c2, c3], 2):
            assert (
                hstrat.calc_rank_of_earliest_detectable_mrca_among(
                    [wrap(x1), wrap(x2)],
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_earliest_detectable_mrca_among(
                    [wrap(x2), wrap(x1)],
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_earliest_detectable_mrca_among(
                    [wrap(x1), wrap(x2), wrap(x1)],
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_earliest_detectable_mrca_among(
                    [wrap(x2), wrap(x1)] * 10,
                    confidence_level=confidence_level,
                )
                is None
            )


@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95, 0.99],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_rank_of_earliest_detectable_mrca_among2(
    confidence_level,
    differentia_bit_width,
    wrap,
):
    expected_thresh = (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_bit_width,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1 - confidence_level,
        )
        - 1
    )

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_bit_width,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    c2 = c1.Clone()
    c3 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_bit_width,
        stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
    )

    for x1, x2 in it.combinations([c1, c2, c3], 2):
        x1 = x1.Clone()
        x2 = x2.Clone()
        while (
            hstrat.get_nth_common_rank_between(x1, x2, expected_thresh) is None
        ):
            random.choice([x1, x2]).DepositStratum()

        assert hstrat.calc_rank_of_earliest_detectable_mrca_among(
            [wrap(x1), wrap(x2)],
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x1, x2, expected_thresh)
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among(
            [wrap(x2), wrap(x1)],
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x2, x1, expected_thresh)
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among(
            [wrap(x1), wrap(x2), wrap(x1)],
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x1, x2, expected_thresh)
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among(
            [wrap(x2), wrap(x1)] * 10,
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x2, x1, expected_thresh)

        for __ in range(3):
            x1.DepositStratum()
            x2.DepositStratum()

        assert hstrat.calc_rank_of_earliest_detectable_mrca_among(
            [wrap(x1), wrap(x2)],
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x1, x2, expected_thresh)
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among(
            [wrap(x2), wrap(x1)],
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x2, x1, expected_thresh)
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among(
            [wrap(x1), wrap(x2), wrap(x1)],
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x1, x2, expected_thresh)
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among(
            [wrap(x2), wrap(x1)] * 10,
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x2, x1, expected_thresh)


@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_rank_of_earliest_detectable_mrca_among3(wrap):

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    for __ in range(10):
        assert (
            hstrat.calc_rank_of_earliest_detectable_mrca_among(
                [wrap(c1), wrap(c1)]
            )
            == 0
        )
        assert (
            hstrat.calc_rank_of_earliest_detectable_mrca_among(
                [wrap(c1), wrap(c1)] * 10
            )
            == 0
        )
        c1.DepositStratum()


@pytest.mark.filterwarnings("ignore:Empty or singleton population.")
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_rank_of_earliest_detectable_mrca_among4(wrap):

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    for __ in range(10):
        assert (
            hstrat.calc_rank_of_earliest_detectable_mrca_among([wrap(c1)])
            is None
        )
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among([]) is None
        c1.DepositStratum()


@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_rank_of_earliest_detectable_mrca_among_generator(wrap):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    for __ in range(10):
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among(
            [wrap(c1) for __ in range(10)]
        ) == hstrat.calc_rank_of_earliest_detectable_mrca_among(
            (wrap(c1) for __ in range(10))
        )
        c1.DepositStratum()

    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    for __ in range(10):
        assert hstrat.calc_rank_of_earliest_detectable_mrca_among(
            [wrap(c2) for __ in range(10)]
        ) == hstrat.calc_rank_of_earliest_detectable_mrca_among(
            (wrap(c2) for __ in range(10))
        )
        c2.DepositStratum()



import random

import pytest

from hstrat._auxiliary_lib import with_rng_state_context


def test_with_rng_state_context_calls_function():
    @with_rng_state_context(seed=123)
    def my_function():
        nonlocal called
        called = True

    called = False
    my_function()
    assert called


def test_with_rng_state_context_sets_seed():
    random.seed(1)
    assert random.getstate() != random.Random(123).getstate()

    @with_rng_state_context(seed=123)
    def my_function():
        assert random.getstate() == random.Random(123).getstate()

    my_function()


def test_with_rng_state_context_works_with_exceptions():
    random.seed(456)
    assert random.getstate() == random.Random(456).getstate()

    @with_rng_state_context(seed=123)
    def my_function():
        assert random.getstate() != random.Random(456).getstate()
        raise ValueError

    with pytest.raises(ValueError):
        my_function()

    assert random.getstate() == random.Random(456).getstate()



import numpy as np

from ...._auxiliary_lib import is_in_unit_test


def _select_parents(
    island_niche_size: int,
    tournament_size: int,
    pop_arr: np.array,
) -> np.array:
    """Perform tournament selection within island-niches.

    The population array `pop_arr` is not altered.

    Returns
    -------
    numpy.array of int
        Population locations (i.e., indices) specifying the parent for each
        population slot at the next generation.

        Position within array corresponds to post-turnover population members'
        population positions. Values within array correspond to those members'
        parents' population positions within the pre-turnover population.
    """
    num_island_niches = len(pop_arr) // island_niche_size

    tournament_boosters = (
        np.broadcast_to(
            np.repeat(
                np.arange(num_island_niches),
                island_niche_size,
            ),
            shape=(tournament_size, pop_arr.size),
        ).T
        * island_niche_size
    )
    tournament_rosters = (
        np.random.randint(
            0,
            island_niche_size,
            (pop_arr.size, tournament_size),
        )
        + tournament_boosters
    )

    tournament_fitnesses = np.take(pop_arr, tournament_rosters)
    winning_tournament_positions = tournament_fitnesses.argmax(axis=1)

    # https://stackoverflow.com/a/61234228
    winning_tournament_locs = np.take_along_axis(
        tournament_rosters,
        winning_tournament_positions[:, None],
        axis=1,
    )

    if is_in_unit_test():
        assert len(winning_tournament_locs.flatten()) == pop_arr.size
        assert np.all(
            np.arange(pop_arr.size) // island_niche_size
            == winning_tournament_locs.flatten() // island_niche_size
        )

    return winning_tournament_locs.flatten()



from collections import deque
import typing

from ...genome_instrumentation import HereditaryStratigraphicColumn
from .._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)


def calc_rank_of_first_retained_disparity_between_generic(
    first: HereditaryStratigraphicColumn,
    second: HereditaryStratigraphicColumn,
    *,
    first_start_idx: int = 0,
    second_start_idx: int = 0,
    confidence_level: float,
) -> typing.Optional[int]:
    """Find first mismatching strata between columns.

    Implementation detail. Provides general-case implementation.
    """
    # helper setup
    try:
        first_iter = first._stratum_ordered_store.IterRankDifferentiaZip(
            get_rank_at_column_index=first.GetRankAtColumnIndex,
            start_column_index=first_start_idx,
        )
        second_iter = second._stratum_ordered_store.IterRankDifferentiaZip(
            get_rank_at_column_index=second.GetRankAtColumnIndex,
            start_column_index=second_start_idx,
        )
    except AttributeError:
        first_iter = zip(
            first.IterRetainedRanks(), first.IterRetainedDifferentia()
        )
        second_iter = zip(
            second.IterRetainedRanks(), second.IterRetainedDifferentia()
        )

    first_cur_rank, first_cur_differentia = next(first_iter)
    second_cur_rank, second_cur_differentia = next(second_iter)
    first_prev_rank: int
    second_prev_rank: int

    def advance_first():
        nonlocal first_prev_rank, first_cur_rank
        nonlocal first_cur_differentia, first_iter
        try:
            first_prev_rank = first_cur_rank
            first_cur_rank, first_cur_differentia = next(first_iter)
            _ = first_cur_differentia
        except StopIteration:
            first_iter = None

    def advance_second():
        nonlocal second_prev_rank, second_cur_rank
        nonlocal second_cur_differentia, second_iter
        try:
            second_prev_rank = second_cur_rank
            second_cur_rank, second_cur_differentia = next(second_iter)
            _ = second_cur_differentia
        except StopIteration:
            second_iter = None

    # we need to keep track of enough last-seen common ranks so that we
    # can discount this many (minus 1) as potentially occuring due to
    # spurious differentia collisions
    assert (
        first.GetStratumDifferentiaBitWidth()
        == second.GetStratumDifferentiaBitWidth()
    )
    collision_implausibility_threshold = calc_min_implausible_spurious_consecutive_differentia_collisions_between(
        first,
        second,
        significance_level=1.0 - confidence_level,
    )
    assert collision_implausibility_threshold > 0
    # holds up to n last-seen common ranks,
    # with the newest last-seen rank at the front (index 0)
    # and the up to nth last-seen rank at the back (index -1)
    preceding_common_ranks = deque([], collision_implausibility_threshold)
    # a.k.a.
    # while (
    #     first_column_idx < first.GetNumStrataRetained()
    #     and second_column_idx < second.GetNumStrataRetained()
    # ):
    while first_iter is not None and second_iter is not None:
        if first_cur_rank == second_cur_rank:
            # strata at same rank can be compared
            if first_cur_differentia == second_cur_differentia:
                preceding_common_ranks.appendleft(first_cur_rank)
                # matching differentiae at the same rank,
                # keep searching for mismatch
                # advance first and second
                # must ensure both advance, even if one stops iteration
                advance_first()
                advance_second()
            else:
                # mismatching differentiae at the same rank
                preceding_common_ranks.appendleft(first_cur_rank)
                assert 0 <= first_cur_rank < first.GetNumStrataDeposited()

                # discount collision_implausibility_threshold - 1 common
                # ranks due to potential spurious differentia collisions;
                # if not enough common ranks are available we still know
                # *definitively* that a disparity occured (because we
                # observed disparite strata at the same rank); so, make the
                # conservative assumption that the disparity occured as far
                # back as possible (the oldest up to nth last-seen common
                # rank, at the back of the deque)
                return preceding_common_ranks[-1]
        elif first_cur_rank < second_cur_rank:
            # current stratum on first column older than on second column
            # advance to next-newer stratum on first column
            advance_first()
        elif first_cur_rank > second_cur_rank:
            # current stratum on second column older than on first column
            # advance to next-newer stratum on second column
            advance_second()

    if first_iter is not None:
        # although no mismatching strata found between first and second
        # first has strata ranks beyond the newest found in second
        # conservatively assume mismatch will be with next rank of second
        assert second_iter is None
        assert second_prev_rank + 1 == min(
            first.GetNumStrataDeposited(), second.GetNumStrataDeposited()
        )
        preceding_common_ranks.appendleft(second_prev_rank + 1)
        res = preceding_common_ranks[-1]
        assert 0 <= res <= first.GetNumStrataDeposited()
        assert 0 <= res <= second.GetNumStrataDeposited()
        return res
    elif second_iter is not None:
        # although no mismatching strata found between second and first
        # second has strata ranks beyond the newest found in first
        # conservatively assume mismatch will be with next rank
        assert first_iter is None
        assert first_prev_rank + 1 == min(
            first.GetNumStrataDeposited(), second.GetNumStrataDeposited()
        )
        preceding_common_ranks.appendleft(first_prev_rank + 1)
        res = preceding_common_ranks[-1]
        assert 0 <= res <= first.GetNumStrataDeposited()
        assert 0 <= res <= second.GetNumStrataDeposited()
        return res
    else:
        # no disparate strata found
        # and first and second have the same newest rank
        assert first_iter is None and second_iter is None
        preceding_common_ranks.appendleft(None)
        assert (
            preceding_common_ranks[-1] is None
            or preceding_common_ranks[-1] >= 0
        )
        return preceding_common_ranks[-1]



import typing

from matplotlib import pyplot as plt

from ..._auxiliary_lib import ScalarFormatterFixedPrecision
from ._mrca_uncertainty_absolute_barplot import (
    mrca_uncertainty_absolute_barplot,
)
from ._mrca_uncertainty_relative_barplot import (
    mrca_uncertainty_relative_barplot,
)
from ._strata_retained_frac_lineplot import strata_retained_frac_lineplot
from ._strata_retained_num_lineplot import strata_retained_num_lineplot
from ._stratum_retention_dripplot import stratum_retention_dripplot


def policy_panel_plot(
    stratum_retention_policy: typing.Any,
    num_generations: int,
    do_show: bool = False,
    fig: typing.Optional[plt.matplotlib.figure.Figure] = None,
) -> plt.matplotlib.figure.Figure:
    """Holisticaly sketch stratum retention policy at a particular generation.

    Produces a composite multipanel plot.

    Parameters
    ----------
    stratum_retention_policy : any
        Object specifying stratum retention policy.
    num_generations : int
        Number of generations to plot.
    do_show : bool, optional
        Whether to show() the plot automatically.
    fig : matplotlib/pylab figure, optional
        If a valid matplotlib.figure.Figure instance, the plot is drawn in that
        Figure. By default (None), a new figure is created.
    """
    if fig is None:
        fig = plt.figure(figsize=(8, 6))
    elif not isinstance(fig, plt.matplotlib.figure.Figure):
        raise ValueError(f"Invalid argument for fig: {fig}")

    top_left_ax = fig.add_subplot(3, 2, 1)
    mid_left_ax = fig.add_subplot(3, 2, 3)
    bot_left_ax = fig.add_subplot(3, 2, 5)
    top_right_ax = fig.add_subplot(2, 2, 2)
    bot_right_ax = fig.add_subplot(2, 2, 4)
    fig.suptitle(str(stratum_retention_policy))

    # left column
    stratum_retention_dripplot(
        stratum_retention_policy=stratum_retention_policy,
        num_generations=num_generations,
        ax=top_left_ax,
        do_show=False,
    )
    mrca_uncertainty_absolute_barplot(
        stratum_retention_policy=stratum_retention_policy,
        num_generations=num_generations,
        ax=mid_left_ax,
        do_show=False,
    )
    mrca_uncertainty_relative_barplot(
        stratum_retention_policy=stratum_retention_policy,
        num_generations=num_generations,
        ax=bot_left_ax,
        do_show=False,
    )

    # right column
    strata_retained_num_lineplot(
        stratum_retention_policy=stratum_retention_policy,
        num_generations=num_generations,
        ax=top_right_ax,
        do_show=False,
    )

    strata_retained_frac_lineplot(
        stratum_retention_policy=stratum_retention_policy,
        num_generations=num_generations,
        ax=bot_right_ax,
        do_show=False,
    )

    # TODO bot_right_ax text panel with params etc

    top_left_ax.get_yaxis().set_label_coords(-0.15, 0.5)
    mid_left_ax.get_yaxis().set_label_coords(-0.15, 0.5)
    bot_left_ax.get_yaxis().set_label_coords(-0.15, 0.5)
    top_right_ax.set(xlabel=None)
    top_left_ax.set(xlabel=None)
    mid_left_ax.set(xlabel=None)
    fig.subplots_adjust(wspace=0.3)
    fig.subplots_adjust(hspace=0.3)

    # can't reuse single object due to side effects between plots
    def make_fixed_prec_sci_formatter():
        formatter = ScalarFormatterFixedPrecision()
        formatter.set_scientific(True)
        formatter.set_powerlimits((0, 0))
        return formatter

    all_axes = (
        top_left_ax,
        mid_left_ax,
        bot_left_ax,
        top_right_ax,
        bot_right_ax,
    )
    for ax in all_axes:
        ax.yaxis.set_major_formatter(make_fixed_prec_sci_formatter())

    # ensure left col vertical alignment
    left_col = top_left_ax, mid_left_ax, bot_left_ax
    left_col_xlim = (
        min((ax.get_xlim()[0] for ax in left_col)),
        max((ax.get_xlim()[1] for ax in left_col)),
    )
    for ax in left_col:
        ax.set_xlim(left_col_xlim)

    # ensure right col vertical alignment
    right_col = top_right_ax, bot_right_ax
    right_col_xlim = (
        min((ax.get_xlim()[0] for ax in right_col)),
        max((ax.get_xlim()[1] for ax in right_col)),
    )
    for ax in right_col:
        ax.set_xlim(right_col_xlim)

    if do_show:
        plt.show()

    return fig



import typing

from ...._auxiliary_lib import HereditaryStratigraphicArtifact
from ....juxtaposition import calc_probability_differentia_collision_between
from ...estimators import (
    estimate_rank_of_mrca_unbiased as estimate_rank_of_mrca_unbiased_,
)
from ._extract_common_retained_ranks_through_first_retained_disparity import (
    extract_common_retained_ranks_through_first_retained_disparity,
)


def estimate_rank_of_mrca_unbiased(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    prior: typing.Any,
) -> typing.Optional[float]:
    """Forwards to estimate_rank_of_mrca_unbiased estimator."""
    waypoints_ascending = (
        extract_common_retained_ranks_through_first_retained_disparity(
            first, second
        )
    )
    return estimate_rank_of_mrca_unbiased_(
        reversed(waypoints_ascending),
        calc_probability_differentia_collision_between(first, second),
        prior,
    )



import itertools as it
import operator
import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_rank_of_earliest_detectable_mrca_among import (
    calc_rank_of_earliest_detectable_mrca_among,
)
from ._calc_rank_of_mrca_bounds_among import calc_rank_of_mrca_bounds_among


def calc_rank_of_mrca_uncertainty_among(
    population: typing.Iterable[HereditaryStratigraphicArtifact],
    prior: str,
    confidence_level: float = 0.95,
) -> typing.Optional[int]:
    """How wide is the estimate window for generation of MRCA?

    Calculate uncertainty of estimate for the number of depositions
    elapsed along the line of descent before the most common recent
    ancestor with second.

    Returns 0 if no common ancestor between first and second can be resolved
    with sufficient confidence. If insufficient common ranks between first
    and second are available to resolve any common ancestor, returns None. If
    population is empty or singleton, also returns None.

    See Also
    --------
    calc_rank_of_mrca_bounds_among :
        Calculates bound whose uncertainty this method reports. See the
        corresponding docstring for explanation of parameters.
    """
    pop_tee1, pop_tee2 = it.tee(population)
    if (
        calc_rank_of_earliest_detectable_mrca_among(
            pop_tee1,
            confidence_level=confidence_level,
        )
        is None
    ):
        return None

    bounds = calc_rank_of_mrca_bounds_among(
        pop_tee2,
        prior=prior,
        confidence_level=confidence_level,
    )
    return 0 if bounds is None else abs(operator.sub(*bounds)) - 1



import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._unfurl_lineage_with_contiguous_ids import (
    unfurl_lineage_with_contiguous_ids,
)


def alifestd_unfurl_lineage_asexual(
    phylogeny_df: pd.DataFrame,
    leaf_id: int,
    mutate: bool = False,
) -> np.array:
    """List `leaf_id` and its ancestor id sequence through tree root.

    The provided dataframe must be asexual.
    """

    phylogeny_df = alifestd_try_add_ancestor_id_col(
        phylogeny_df, mutate=mutate
    )
    if alifestd_has_contiguous_ids(phylogeny_df):
        return unfurl_lineage_with_contiguous_ids(
            phylogeny_df["ancestor_id"].to_numpy(),
            leaf_id,
        )
    else:
        ancestor_lookup = dict(
            zip(phylogeny_df["id"], phylogeny_df["ancestor_id"])
        )
        cur_id = leaf_id
        res = []
        while True:
            res.append(cur_id)
            next_id = ancestor_lookup[cur_id]
            if next_id == cur_id:
                break
            cur_id = next_id

        return np.array(res)



from iterpop import iterpop as ip
import numpy as np
import pytest

from hstrat.hstrat import recency_proportional_resolution_curbed_algo


@pytest.mark.parametrize(
    "size_curb",
    [
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10,
        ),
        (2**32,),
    ],
)
def test_zero_and_last_ranks_retained(size_curb, time_sequence):
    policy = recency_proportional_resolution_curbed_algo.Policy(size_curb)
    spec = policy.GetSpec()
    instance = recency_proportional_resolution_curbed_algo.IterRetainedRanks(
        spec
    )
    for num_strata_deposited in time_sequence:
        for which in (
            instance,
            recency_proportional_resolution_curbed_algo.IterRetainedRanks(
                spec
            ),
        ):
            res = which(
                policy,
                num_strata_deposited,
            )
            if num_strata_deposited > 1:
                first, *_middle, last = res
                assert first == 0
                assert last == num_strata_deposited - 1
            elif num_strata_deposited == 1:
                assert ip.popsingleton(res) == 0
            else:
                assert next(res, None) is None



"""Retain as few strata as possible.

The nominal resolution policy only retains the most ancient (i.e., very
first) and most recent (i.e., last) strata. So, comparisons between two
columns under this policy will only be able to detect whether they share
any common ancestor and whether they are from the same organism (i.e., no
generations have elapsed since the MRCA). Thus, MRCA rank estimate
uncertainty scales as O(n) with respect to the greater number of strata deposited on either column.

Under the nominal resolution policy, the number of strata retained (i.e.,
space complexity) scales as O(1) with respect to the number of strata
deposited.
"""

from ._Policy import Policy
from ._PolicySpec import PolicySpec
from ._enact._GenDropRanks import GenDropRanks
from ._enact._GenDropRanks_ import impls as _GenDropRanks_impls
from ._invar._CalcMrcaUncertaintyAbsUpperBound import (
    CalcMrcaUncertaintyAbsUpperBound,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBound import (
    CalcMrcaUncertaintyRelUpperBound,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundPessimalRank,
)
from ._invar._CalcNumStrataRetainedUpperBound import (
    CalcNumStrataRetainedUpperBound,
)
from ._scry._CalcMrcaUncertaintyAbsExact import CalcMrcaUncertaintyAbsExact
from ._scry._CalcMrcaUncertaintyRelExact import CalcMrcaUncertaintyRelExact
from ._scry._CalcNumStrataRetainedExact import CalcNumStrataRetainedExact
from ._scry._CalcRankAtColumnIndex import CalcRankAtColumnIndex
from ._scry._IterRetainedRanks import IterRetainedRanks

__all__ = [
    "CalcMrcaUncertaintyAbsExact",
    "CalcMrcaUncertaintyAbsUpperBound",
    "CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyAbsUpperBoundPessimalRank",
    "CalcMrcaUncertaintyRelExact",
    "CalcMrcaUncertaintyRelUpperBound",
    "CalcMrcaUncertaintyRelUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyRelUpperBoundPessimalRank",
    "CalcNumStrataRetainedExact",
    "CalcNumStrataRetainedUpperBound",
    "CalcRankAtColumnIndex",
    "GenDropRanks",
    "_GenDropRanks_impls",
    "IterRetainedRanks",
    "Policy",
    "PolicySpec",
]

from ...._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



r"""Provide log space complexity and recency-proportional strata spacing.

The MRCA-recency-proportional resolution policy ensures estimates of MRCA
rank will have uncertainty bounds less than or equal to a user-specified
proportion of the actual number of generations elapsed since the MRCA and the
deepest of the compared columns. MRCA rank estimate uncertainty scales in the
worst case scales as O(n) with respect to the greater number of strata
deposited on either column. However, with respect to estimating the rank of the
MRCA when lineages diverged any fixed number of generations ago, uncertainty
scales as O(1).

Under the MRCA-recency-proportional resolution policy, the number of strata
retained (i.e., space complexity) scales as O(log(n)) with respect to the
number of strata deposited.


This functor's retention policy implementation guarantees that columns will
retain appropriate strata so that for any two columns with m and n strata
deposited, the rank of the most recent common ancestor at rank k can be
determined with uncertainty of at most

    bound = floor(
        max(m - k, n - k)
        / recency_proportional_resolution
    )

ranks.

How does the predicate work and how does it guarantee this resolution?

To begin, let's consider setting up just the *first* rank of the stratum after
the root ancestor we will retain.

root ancestor                                     extant individual
|                                                                 |
|                    num_strata_deposited                         |
| ___________________________/\_________________________________  |
|/                                                               \|
|-------------------|#############################################|
 \_______  ________/|\____________________  ______________________/
         \/         |                     \/
   max_uncertainty  |            worst_case_mrca_depth
                    |
                    proposed retention rank

To provide guaranteed resolution, max_uncertainty must be leq than

    worst_case_mrca_depth // guaranteed_resolution

So, to find the largest permissible max_uncertainty we must solve

    max_uncertainty = worst_case_mrca_depth // guaranteed_resolution

By construction we have

    worst_case_mrca_depth = num_strata_deposited - max_uncertainty

Substituting into the above expression gives

    max_uncertainty
    = (num_strata_deposited - max_uncertainty) // guaranteed_resolution

Solving for max_uncertainty yields

    max_uncertainty
    = num_strata_deposited // (guaranteed_resolution + 1)

We now have an upper bound for the rank of the first stratum rank we must
retain. We can repeat this process recursively to select ranks that give
adequate resolution proportional to worst_case_mrca_depth.

However, we must guarantee that thes ranks are actually available for us to
retain (i.e., it hasn't been purged out of the column at a previous time point
as the column was grown by successive deposition). We will do this by picking
the rank that is the highest power of 2 less than or equal to our bound. If we
repeat this procedure as we recurse, we are guaranteed that this rank will have
been preserved across all previous timepoints.

This is because a partial sum sequence where all elements are powers of 2 and
elements in the sequence are will include all multiples of powers of 2 greater
than or equal to the first element that are less than or equal to the sum of
the entire sequence.

An example is the best way to convince yourself. Thinking analogously in base
10,

    100 + 10... + 1...

the partial sums of any sequence of this form will always include all multiples
of powers of 100, 1000, etc. that are less than or equal to the sum of the
entire sequence.

In our application, partial sums represent retained ranks. So, all ranks that
are perfect powers of 2 measuring from the root ancestor will have been
retained after being deposited. This property generalizes recursively.
"""

from ._Policy import Policy
from ._PolicySpec import PolicySpec
from ._enact._GenDropRanks import GenDropRanks
from ._enact._GenDropRanks_ import impls as _GenDropRanks_impls
from ._invar._CalcMrcaUncertaintyAbsUpperBound import (
    CalcMrcaUncertaintyAbsUpperBound,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBound import (
    CalcMrcaUncertaintyRelUpperBound,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundPessimalRank,
)
from ._invar._CalcNumStrataRetainedUpperBound import (
    CalcNumStrataRetainedUpperBound,
)
from ._scry._CalcMrcaUncertaintyAbsExact import CalcMrcaUncertaintyAbsExact
from ._scry._CalcMrcaUncertaintyRelExact import CalcMrcaUncertaintyRelExact
from ._scry._CalcNumStrataRetainedExact import CalcNumStrataRetainedExact
from ._scry._CalcRankAtColumnIndex import CalcRankAtColumnIndex
from ._scry._IterRetainedRanks import IterRetainedRanks

__all__ = [
    "CalcMrcaUncertaintyAbsExact",
    "CalcMrcaUncertaintyAbsUpperBound",
    "CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyAbsUpperBoundPessimalRank",
    "CalcMrcaUncertaintyRelExact",
    "CalcMrcaUncertaintyRelUpperBound",
    "CalcMrcaUncertaintyRelUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyRelUpperBoundPessimalRank",
    "CalcNumStrataRetainedExact",
    "CalcNumStrataRetainedUpperBound",
    "CalcRankAtColumnIndex",
    "GenDropRanks",
    "_GenDropRanks_impls",
    "IterRetainedRanks",
    "Policy",
    "PolicySpec",
]

from ...._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import pandas as pd

from ._alifestd_find_root_ids import alifestd_find_root_ids
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids


def alifestd_mark_roots(
    phylogeny_df: pd.DataFrame, mutate: bool = False
) -> pd.DataFrame:
    """Create column `is_root` to mark rows with no ancestor.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    root_ids = alifestd_find_root_ids(phylogeny_df)
    phylogeny_df["is_root"] = False
    phylogeny_df.loc[root_ids, "is_root"] = True

    return phylogeny_df



import unittest

from hstrat._auxiliary_lib import bit_floor


class TestBitFloor(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test(self):
        assert bit_floor(0b00000000) == 0b00000000
        assert bit_floor(0b00000001) == 0b00000001
        assert bit_floor(0b00000010) == 0b00000010
        assert bit_floor(0b00000011) == 0b00000010
        assert bit_floor(0b00000100) == 0b00000100
        assert bit_floor(0b00000101) == 0b00000100
        assert bit_floor(0b00000110) == 0b00000100
        assert bit_floor(0b00000111) == 0b00000100
        assert bit_floor(0b00001000) == 0b00001000
        assert bit_floor(0b00001001) == 0b00001000


if __name__ == "__main__":
    unittest.main()



import numpy as np

from ._jit import jit


@jit(nopython=True)
def all_unique(array: np.array) -> bool:
    """Are all values in `array` unique?"""
    seen_unique = set()
    for idx, val in enumerate(array):
        seen_unique.add(val)
        if idx + 1 != len(seen_unique):
            return False

    return True



import typing

import numpy as np


def to_tril(matrix: np.ndarray) -> typing.List[typing.List[float]]:
    """Convert a distance matrix to the lower triangular form representation
    required by BioPython.

    Excludes upper triangular entries (i.e., returned list of lists is ragged).
    Sets all diagonal elements to zero.

    Parameters
    ----------
    matrix : np.ndarray
        The input matrix, represented as a 2-dimensional numpy array.

    Returns
    -------
    typing.List[typing.List[float]]
        The lower triangular form of the input matrix, represented as a list of
        lists of floats.

    Examples
    --------
    >>> import numpy as np
    >>> matrix = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> to_tril(matrix)
    [[0.0], [4, 0.0, 0], [7, 8, 0.0]]
    """
    return [
        row[:row_idx] + [0.0] for row_idx, row in enumerate(matrix.tolist())
    ]



import typing

from ..._auxiliary_lib import demark
from ...genome_instrumentation import HereditaryStratigraphicColumn
from ._descend_template_phylogeny_naive import descend_template_phylogeny_naive
from ._descend_template_phylogeny_posthoc import (
    descend_template_phylogeny_posthoc,
)


def descend_template_phylogeny(
    ascending_lineage_iterables: typing.Iterable[typing.Iterable],
    descending_tree_iterable: typing.Iterable,
    get_parent: typing.Callable[[typing.Any], typing.Any],
    get_stem_length: typing.Callable[[typing.Any], int],
    seed_column: HereditaryStratigraphicColumn,
    demark: typing.Callable[[typing.Any], typing.Hashable] = demark,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.List[HereditaryStratigraphicColumn]:
    """Generate a population of hereditary stratigraphic columns that could
    have resulted from the template phylogeny.

    Parameters
    ----------
    ascending_lineage_iterables : iterable of phylogeny node iterables
        Iterable that yields an acending lineage iterable for each extant
        individual from template phylogeny.

        The acending lineage iterable should yield the extant individual's
        phylogeny node object and then successive parent node objects until
        yielding the root node object.
    descending_tree_iterable : iterable of phylogeny nodes
        Iterable over all phylogeny node objects in template phylogeny.

        The root node object must be yielded first. Then, each phylogeny node
        object should be yielded, with no object being yielded before its
        parent object. Any topologically sorted ordering is allowed, including
        pre-order depth-first iteration and breadth-first (a.k.a. leve-order)
        iteration.
    get_parent : function
        Function that returns the parent node object for a phylogeny node
        object.

        Will never be called on root node object.
    get_stem_length : function
        Function that returns the length of the subtending edge for a phylogeny
        node.

        Length should be an integer value representing the number of
        generations elapsed between the focal node's parent and the focal node.
        Will never be called on root node object.

        For phylogenies including all ancestors as distinct nodes, `lambda
        node: 1` can be specified.
    seed_column : HereditaryStratigraphicColumn
        Hereditary stratigraphic column to seed at root node of phylogeny.

        Returned hereditary stratigraphic column population will be generated
        as if repeatedly calling `CloneDescendant()` on `seed_column`. As such,
        specifies configuration (i.e., differentia bit width and stratum
        retention policy) for returned columns. May already have strata
        deposited, which will be incorporated into generated extant population.
    demark : function, default _auxiliary_lib.demark
        Function that converts returned nodes to a unique hashable value.

        For object-based Nodes, this might be `id`. For value-based Nodes, this
        might be an identity function.

        If default, will use runtime type information to make a reasonable
        guess. Passing a custom value allows bypass of this slow type
        inspection.
    progress_wrap : Callable, default identity function
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.

    Returns
    -------
    list of HereditaryStratigraphicColumn
        Population of hereditary stratigraphic columns for extant lineage
        members.

        Order of columns corresponds to order of `ascending_lineage_iterables`.

    Notes
    -----
    To improve performance, consider preprocessing template phylogeny
    representation to collapse unifurcation chains (i.e., converting chains of
    `n` nodes to a single node with stem length `n`).

    Delegates to `descend_template_phylogeny_naive` or
    `descend_template_phylogeny_posthoc`, depending on wheter retained ranks
    at an arbitrary generation can be calculated by the `seed_column`'s stratum
    retention policy.
    """

    impl = (
        descend_template_phylogeny_posthoc
        if seed_column._stratum_retention_policy.IterRetainedRanks is not None
        else descend_template_phylogeny_naive
    )

    return impl(
        ascending_lineage_iterables,
        descending_tree_iterable,
        get_parent,
        get_stem_length,
        seed_column,
        demark=demark,
        progress_wrap=progress_wrap,
    )



from hstrat._auxiliary_lib import popcount


def test_popcount():
    assert popcount(0) == 0
    assert popcount(1) == 1
    assert popcount(2) == 1
    assert popcount(3) == 2
    assert popcount(4) == 1
    assert popcount(5) == 2
    assert popcount(6) == 2
    assert popcount(7) == 3
    assert popcount(12481027492184092184390812949318513097509831) == 76



import pytest

from hstrat.hstrat import recency_proportional_resolution_curbed_algo

from ..impl import iter_backing_policy_transition_ranks


@pytest.mark.parametrize(
    "size_curb",
    [
        1,
        2,
        3,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        42,
        97,
        100,
        254,
        255,
        256,
        257,
    ],
)
def test_backing_policy_transition_consistency(size_curb):
    policy = recency_proportional_resolution_curbed_algo.Policy(size_curb)
    spec = policy.GetSpec()
    instance = recency_proportional_resolution_curbed_algo.IterRetainedRanks(
        spec
    )

    for num_strata_deposited in (
        test_rank
        for transition_rank in iter_backing_policy_transition_ranks(size_curb)
        for test_rank in range(
            max(transition_rank - 2, 0), transition_rank + 3
        )
    ):
        for which in (
            instance,
            recency_proportional_resolution_curbed_algo.IterRetainedRanks(
                spec
            ),
        ):
            cur_set = {
                *which(
                    policy,
                    num_strata_deposited,
                )
            }
            next_set = {
                *which(
                    policy,
                    num_strata_deposited + 1,
                )
            }
            assert (next_set - {num_strata_deposited}) - cur_set == set()
            assert cur_set.issuperset(next_set - {num_strata_deposited})



import typing

import numpy as np
import pandas as pd

from ._alifestd_has_compact_ids import alifestd_has_compact_ids
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._jit import jit


@jit(nopython=True)
def _alifestd_find_chronological_inconsistency_contiguous(
    ancestor_ids: np.array, origin_times: np.array
) -> typing.Optional[int]:
    for id_, ancestor_id in enumerate(ancestor_ids):
        # self comparison ok for geneses
        if origin_times[ancestor_id] > origin_times[id_]:
            return id_
    return None


@jit(nopython=True)
def _alifestd_find_chronological_inconsistency_compact(
    ids: np.array, ancestor_ids: np.array, origin_times: np.array
) -> typing.Optional[int]:
    origin_time_lookup = np.empty(len(ids))
    for (id_, origin_time) in zip(ids, origin_times):
        origin_time_lookup[id_] = origin_time

    for id_, ancestor_id in zip(ids, ancestor_ids):
        # self comparison ok for geneses
        if origin_time_lookup[ancestor_id] > origin_time_lookup[id_]:
            return id_
    return None


@jit(nopython=True)
def _alifestd_find_chronological_inconsistency(
    ids: np.array, ancestor_ids: np.array, origin_times: np.array
) -> typing.Optional[int]:
    origin_time_lookup = dict()
    for (id_, origin_time) in zip(ids, origin_times):
        origin_time_lookup[id_] = origin_time

    for id_, ancestor_id in zip(ids, ancestor_ids):
        # self comparison ok for geneses
        if origin_time_lookup[ancestor_id] > origin_time_lookup[id_]:
            return id_
    return None


def alifestd_find_chronological_inconsistency(
    phylogeny_df: pd.DataFrame,
) -> typing.Optional[int]:
    """Return the id of a taxon with origin time precedint its parent's, if
    any are present."""

    if "ancestor_id" in phylogeny_df:
        if alifestd_has_contiguous_ids(phylogeny_df):
            return _alifestd_find_chronological_inconsistency_contiguous(
                phylogeny_df["ancestor_id"].to_numpy(),
                phylogeny_df["origin_time"].to_numpy(),
            )
        elif alifestd_has_compact_ids(phylogeny_df):
            return _alifestd_find_chronological_inconsistency_compact(
                phylogeny_df["id"].to_numpy(),
                phylogeny_df["ancestor_id"].to_numpy(),
                phylogeny_df["origin_time"].to_numpy(),
            )
        else:
            return _alifestd_find_chronological_inconsistency(
                phylogeny_df["id"].to_numpy(),
                phylogeny_df["ancestor_id"].to_numpy(),
                phylogeny_df["origin_time"].to_numpy(),
            )

    phylogeny_df = phylogeny_df.set_index("id", drop=False)

    for _pos, (_idx, row) in enumerate(phylogeny_df.iterrows()):
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            ancestor_loc = phylogeny_df.index.get_loc(ancestor_id)
            if (
                phylogeny_df.iloc[ancestor_loc]["origin_time"]
                > row["origin_time"]
            ):
                return _idx

    return None



import typing

import pandas as pd


class CopyableSeriesItemsIter:
    """An iterator that returns (index, value) pairs from a Pandas series.

    Parameters
    ----------
    series : pd.Series
        The series to iterate over.

    Returns
    -------
    CopyableSeriesItemsIter
        An iterator object that returns (index, value) pairs from the series.

    Raises
    ------
    StopIteration
        If there are no more items in the series to iterate over.

    Examples
    --------
    >>> s = pd.Series([1, 2, 3], index=['a', 'b', 'c'])
    >>> c = CopyableSeriesItemsIter(s)
    >>> list(c)
    [('a', 1), ('b', 2), ('c', 3)]
    """

    _position: int
    _series: pd.Series

    def __init__(self: "CopyableSeriesItemsIter", series: pd.Series) -> None:
        """Initialize a new CopyableSeriesItemsIter instance.

        Parameters
        ----------
        series : pd.Series
            The series to iterate over.
        """
        self._series = series
        self._position = 0

    def __iter__(self: "CopyableSeriesItemsIter") -> "CopyableSeriesItemsIter":
        """Enable iterability.

        Returns
        -------
        CopyableSeriesItemsIter
            The iterator object itself.
        """
        return self

    def __next__(self: "CopyableSeriesItemsIter") -> typing.Tuple:
        """Return the next (index, value) pair from the series.

        Returns
        -------
        typing.Tuple
            A tuple containing the index label and value of the next item in
            the series.

        Raises
        ------
        StopIteration
            If there are no more items in the series to iterate over.
        """
        if self._position >= len(self._series):
            raise StopIteration

        self._position += 1
        return (
            self._series.index[self._position - 1],
            self._series.iloc[self._position - 1],
        )



import math
import typing

import opytional as opyt

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._estimate_rank_of_mrca_between import estimate_rank_of_mrca_between


def estimate_ranks_since_mrca_with(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
    estimator: str,
    prior: typing.Union[str, typing.Any],
) -> typing.Optional[float]:
    """How many generations have elapsed since focal's most recent common
    ancestor with other?

    More specifically, estimate the number of depositions elapsed along focal
    column's line of descent since the most recent common ancestor with other.

    Parameters
    ----------
    estimator : {"maximum_likelihood", "unbiased"}
        What estimation method should be used? Options are "maximum_likelihood"
        or "unbiased".

        See `estimate_ranks_since_mrca_with` for discussion of estimator
        options.
    prior : {"arbitrary", "uniform"} or object implementing prior interface

        Prior probability density distribution over possible generations of the
        MRCA.

        See `estimate_ranks_since_mrca_with` for discussion of prior
        options.

    Returns
    -------
    float, optional
        Estimate of generations elapsed since MRCA, unless first and second
        definitively share no common ancestor in which case None will be
        returned.

    See Also
    --------
    calc_ranks_since_mrca_bounds_with :
        Calculates confidence intervals for generations elapsed along one
        column's line of descent since most recent common ancestor with another
        column.
    does_definitively_have_no_common_anestor :
        Does the hereditary stratigraphic record definitively prove that first
        and second could not possibly share a common ancestor?
    """

    est_rank_of_mrca_between = estimate_rank_of_mrca_between(
        focal,
        other,
        estimator=estimator,
        prior=prior,
    )
    res = opyt.apply_if(
        est_rank_of_mrca_between,
        lambda est: focal.GetNumStrataDeposited()
        - 1
        - est_rank_of_mrca_between,
    )
    if res is not None:
        assert 0 <= res or math.isclose(0, res, abs_tol=10e-6)
        assert res <= focal.GetNumStrataDeposited() or math.isclose(
            res, focal.GetNumStrataDeposited()
        )
    return res



import itertools as it
import typing

import pandas as pd

from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._col_to_dataframe import col_to_dataframe


def pop_to_dataframe(
    columns: typing.Iterable[HereditaryStratigraphicColumn],
    progress_wrap: typing.Callable = lambda x: x,
) -> pd.DataFrame:
    """Create a pandas `DataFrame` summarizing several columns, with retained
    strata as rows.

    Parameters
    ----------
    columns : iterable of HereditaryStratigraphicColumn
        Data to serialize.
    progress_wrap : Callable, default identity function
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.
    """
    return (
        pd.concat(
            map(col_to_dataframe, progress_wrap(columns)),
            keys=it.count(),  # create multiindex, with outer as column id
        )
        .reset_index(  # drop original row-level index
            level=1,
            drop=True,
        )
        .rename_axis(  # set name for column id index
            ["column"],
        )
        .reset_index()
    )



"""Retain all strata to ensure perfect resolution for MRCA estimation.

The perfect resolution policy retains all strata. So, comparisons between
two columns under this policy will detect MRCA rank with zero
uncertainty. So, MRCA rank estimate uncertainty scales as O(1) with respect
to the greater number of strata deposited on either column.

Under the perfect resolution policy, the number of strata retained (i.e.,
space complexity) scales as O(n) with respect to the number of strata
deposited.
"""

from ._Policy import Policy
from ._PolicySpec import PolicySpec
from ._enact._GenDropRanks import GenDropRanks
from ._enact._GenDropRanks_ import impls as _GenDropRanks_impls
from ._invar._CalcMrcaUncertaintyAbsUpperBound import (
    CalcMrcaUncertaintyAbsUpperBound,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBound import (
    CalcMrcaUncertaintyRelUpperBound,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundPessimalRank,
)
from ._invar._CalcNumStrataRetainedUpperBound import (
    CalcNumStrataRetainedUpperBound,
)
from ._scry._CalcMrcaUncertaintyAbsExact import CalcMrcaUncertaintyAbsExact
from ._scry._CalcMrcaUncertaintyRelExact import CalcMrcaUncertaintyRelExact
from ._scry._CalcNumStrataRetainedExact import CalcNumStrataRetainedExact
from ._scry._CalcRankAtColumnIndex import CalcRankAtColumnIndex
from ._scry._IterRetainedRanks import IterRetainedRanks

__all__ = [
    "CalcMrcaUncertaintyAbsExact",
    "CalcMrcaUncertaintyAbsUpperBound",
    "CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyAbsUpperBoundPessimalRank",
    "CalcMrcaUncertaintyRelExact",
    "CalcMrcaUncertaintyRelUpperBound",
    "CalcMrcaUncertaintyRelUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyRelUpperBoundPessimalRank",
    "CalcNumStrataRetainedExact",
    "CalcNumStrataRetainedUpperBound",
    "CalcRankAtColumnIndex",
    "GenDropRanks",
    "_GenDropRanks_impls",
    "IterRetainedRanks",
    "Policy",
    "PolicySpec",
]

from ...._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



"""Provides numba integr array types, or null fallback.

Type must be declared outside jit'ed function or numba fails.
"""


from ._is_in_coverage_run import is_in_coverage_run

try:
    import numba as nb
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    jit_numba_int8_arr_t = None
    jit_numba_int16_arr_t = None
    jit_numba_int32_arr_t = None
    jit_numba_int64_arr_t = None
    jit_numba_uint8_arr_t = None
    jit_numba_uint16_arr_t = None
    jit_numba_uint32_arr_t = None
    jit_numba_uint64_arr_t = None
else:
    if is_in_coverage_run():
        # when numba disabled for coverage testing,
        # using nb.types.int64_ causes numpy TypeError
        jit_numba_int8_arr_t = None
        jit_numba_int16_arr_t = None
        jit_numba_int32_arr_t = None
        jit_numba_int64_arr_t = None
        jit_numba_uint8_arr_t = None
        jit_numba_uint16_arr_t = None
        jit_numba_uint32_arr_t = None
        jit_numba_uint64_arr_t = None
    else:  # pragma: no cover
        jit_numba_int8_arr_t = nb.types.int8[:]
        jit_numba_int16_arr_t = nb.types.int16[:]
        jit_numba_int32_arr_t = nb.types.int32[:]
        jit_numba_int64_arr_t = nb.types.int64[:]
        jit_numba_uint8_arr_t = nb.types.uint8[:]
        jit_numba_uint16_arr_t = nb.types.uint16[:]
        jit_numba_uint32_arr_t = nb.types.uint32[:]
        jit_numba_uint64_arr_t = nb.types.uint64[:]



import typing

from ._pairwise import pairwise


def omit_last(it: typing.Iterable) -> typing.Iterable:
    """Yield iterable values in order, stopping iteration after second-to-last
    value yielded.

    Conceptually equivalent to `[:-1]` slice.
    """
    for first, __ in pairwise(it):
        yield first



import typing

from .._auxiliary_lib import get_hstrat_version


def policy_to_records(policy: typing.Callable) -> typing.Dict:
    """Serialize a stratum retention policy to a dict composed of builtin
    types."""

    spec = policy.GetSpec()
    return {
        "policy_algo": spec.GetAlgoIdentifier(),
        "policy_spec": {
            k.lstrip("_"): v  # strip leading underscores on private members
            for k, v in spec.__dict__.items()
        },
        "policy": policy.GetEvalCtor(),
        "hstrat_version": get_hstrat_version(),
    }



import itertools as it

import numpy as np
import pytest

from hstrat.hstrat import nominal_resolution_algo


@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**2),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=10**3,
            size=10**2,
        ),
    ],
)
def test_policy_consistency(time_sequence):
    policy = nominal_resolution_algo.Policy()
    spec = policy.GetSpec()
    instance = nominal_resolution_algo.CalcMrcaUncertaintyAbsExact(spec)
    for num_strata_deposited in time_sequence:
        retained_ranks = np.fromiter(
            policy.IterRetainedRanks(num_strata_deposited),
            int,
        )
        for actual_mrca_rank in range(num_strata_deposited):
            last_known_commonality = retained_ranks[
                retained_ranks <= actual_mrca_rank,
            ].max(
                initial=0,
            )
            first_known_disparity = retained_ranks[
                retained_ranks > actual_mrca_rank,
            ].min(
                initial=num_strata_deposited,
            )
            policy_requirement = (
                first_known_disparity - last_known_commonality - 1
            )
            assert policy_requirement >= 0
            for which in (
                instance,
                nominal_resolution_algo.CalcMrcaUncertaintyAbsExact(spec),
            ):
                assert (
                    which(
                        policy,
                        num_strata_deposited,
                        num_strata_deposited,
                        actual_mrca_rank,
                    )
                    == policy_requirement
                )


def test_policy_consistency_uneven_branches():
    policy = nominal_resolution_algo.Policy()
    spec = policy.GetSpec()
    instance = nominal_resolution_algo.CalcMrcaUncertaintyAbsExact(spec)
    sample_durations = it.chain(
        range(10**2),
        np.logspace(7, 16, num=10, base=2, dtype="int"),
    )
    for num_strata_deposited_a in sample_durations:
        ranks_a = set(policy.IterRetainedRanks(num_strata_deposited_a))
        for num_strata_deposited_b in sample_durations:
            ranks_b = set(policy.IterRetainedRanks(num_strata_deposited_b))
            retained_ranks = np.fromiter(ranks_a & ranks_b, int)
            least_num_strata_deposited = min(
                num_strata_deposited_a,
                num_strata_deposited_b,
            )
            for actual_mrca_rank in range(least_num_strata_deposited):
                last_known_commonality = retained_ranks[
                    retained_ranks <= actual_mrca_rank,
                ].max(
                    initial=0,
                )
                first_known_disparity = retained_ranks[
                    retained_ranks > actual_mrca_rank,
                ].min(
                    initial=least_num_strata_deposited,
                )
                policy_requirement = (
                    first_known_disparity - last_known_commonality - 1
                )
                assert policy_requirement >= 0
                for which in (
                    instance,
                    nominal_resolution_algo.CalcMrcaUncertaintyAbsExact(spec),
                ):
                    assert (
                        which(
                            policy,
                            num_strata_deposited_a,
                            num_strata_deposited_b,
                            actual_mrca_rank,
                        )
                        == policy_requirement
                    )


def test_eq():
    policy = nominal_resolution_algo.Policy()
    spec = policy.GetSpec()
    instance = nominal_resolution_algo.CalcMrcaUncertaintyAbsExact(spec)

    assert instance == instance
    assert instance == nominal_resolution_algo.CalcMrcaUncertaintyAbsExact(
        spec
    )
    assert instance is not None


def test_negative_index():
    policy = nominal_resolution_algo.Policy()
    spec = policy.GetSpec()
    instance = nominal_resolution_algo.CalcMrcaUncertaintyAbsExact(spec)

    for diff in range(1, 100):
        assert instance(policy, 100, 100, -diff,) == instance(
            policy,
            100,
            100,
            99 - diff,
        )

        assert instance(policy, 101, 100, -diff,) == instance(
            policy,
            101,
            100,
            99 - diff,
        )

        assert instance(policy, 150, 100, -diff,) == instance(
            policy,
            150,
            100,
            99 - diff,
        )

        assert instance(policy, 100, 101, -diff,) == instance(
            policy,
            101,
            100,
            99 - diff,
        )

        assert instance(policy, 100, 150, -diff,) == instance(
            policy,
            150,
            100,
            99 - diff,
        )



from collections import deque
import typing

import anytree


class AnyTreeFastPostOrderIter:
    r"""Iterator that traverses an `anytree.AnyNode` tree using post-order
    strategy (children, self).

    Parameters
    ----------
    root : anytree.AnyNode
        The root node of the tree to iterate over.

    Examples
    --------
    Given the following tree:

        a
       / \
      b   c
     / \   \
    d   e   f

    We can create an instance of `AnyTreeFastPostOrderIter` and use it to
    iterate over the tree:

    >>> from anytree import Node
    >>> root = Node("a")
    >>> b = Node("b", parent=root)
    >>> c = Node("c", parent=root)
    >>> d = Node("d", parent=b)
    >>> e = Node("e", parent=b)
    >>> f = Node("f", parent=c)
    >>> for node in AnyTreeFastPostOrderIter(root):
    ...     print(node.name)
    ...
    d
    e
    b
    f
    c
    a

    Notes
    -----
    This implementationis faster than the `PostOrderIter` implementation provided by `anytree` (especially for large trees). Because it is iterative instead of recursive, it will won't cause recursion limit errors.
    """

    _stack: typing.Deque[anytree.AnyNode]
    _visited: typing.Set[anytree.AnyNode]
    _current: anytree.AnyNode
    _next_node: typing.Optional[anytree.AnyNode]

    def __init__(
        self: "AnyTreeFastPostOrderIter", root: anytree.AnyNode
    ) -> None:
        """Initialize the iterator.

        Parameters
        ----------
        root : anytree.AnyNode
            The root node of the tree to iterate over.
        """
        self._stack = deque()
        self._visited = set()
        self._current = root
        self._next_node = None
        self._traverse_leftmost_path()

    def _traverse_leftmost_path(self: "AnyTreeFastPostOrderIter"):
        """Traverse the leftmost path of the current node, adding visited nodes
        to the stack and setting `_next_node` to the current node when we reach
        the end of the path.

        Also updates the `_visited` set to avoid revisiting nodes.
        """
        while self._current is not None:
            self._stack.append(self._current)
            for child in self._current.children:
                if child not in self._visited:
                    self._current = child
                    break
            else:
                self._next_node = self._current
                self._visited.add(self._current)
                self._current = None

    def __iter__(
        self: "AnyTreeFastPostOrderIter",
    ) -> "AnyTreeFastPostOrderIter":
        """Return the iterator object itself.

        Returns
        -------
        AnyTreeFastPostOrderIter
            The iterator object itself.
        """
        return self

    def __next__(self: "AnyTreeFastPostOrderIter") -> anytree.AnyNode:
        """Return the next node in post-order traversal.

        Returns
        -------
        anytree.AnyNode
            The next node in post-order traversal.

        Raises
        ------
        StopIteration
            When there are no more nodes to traverse.
        """
        if self._next_node is None:
            raise StopIteration
        node = self._next_node
        self._next_node = None
        if len(self._stack) == 0:
            return node
        parent = self._stack.pop()
        if parent.parent is not None and parent.parent not in self._visited:
            self._current = parent.parent
            self._traverse_leftmost_path()
        return node



from base64 import b64decode
import typing

from ._unpack_differentiae_bytes import unpack_differentiae_bytes


def unpack_differentiae_str(
    packed_differentiae: str,
    differentia_bit_width: int,
    num_packed_differentia: typing.Optional[int] = None,
) -> typing.Iterator[int]:
    """Unpack a compact, concatenated base 64 representation into
    a sequence with each element represented as a distinct integer.

    Notes
    -----
    Specifying num_packed_differentia implies that packed_differentiae has
    no header specifying num padding bits.
    """
    bytes_ = b64decode(packed_differentiae)
    yield from unpack_differentiae_bytes(
        bytes_,
        differentia_bit_width,
        num_packed_differentia,
    )



import pandas as pd

from ...._auxiliary_lib import alifestd_find_root_ids


def append_genesis_organism(
    alifestd_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add an entry representing the ancestor of all life organism at
    origin_time 0."""

    if not mutate:
        alifestd_df = alifestd_df.copy()

    if len(alifestd_df):
        assert alifestd_df["origin_time"].min() >= 0
        if alifestd_df["origin_time"].min() == 0:
            return alifestd_df

        genesis_id = alifestd_df["id"].max() + 1
        for old_root_id in alifestd_find_root_ids(alifestd_df):
            if "ancestor_id" in alifestd_df:
                alifestd_df.loc[
                    alifestd_df["id"] == old_root_id, "ancestor_id"
                ] = genesis_id
            alifestd_df.loc[
                alifestd_df["id"] == old_root_id, "ancestor_list"
            ] = f"[{genesis_id}]"
    else:
        genesis_id = 0

    alifestd_df = pd.concat(
        [
            alifestd_df,
            pd.DataFrame(
                {
                    "id": [genesis_id],
                    "ancestor_id": [genesis_id],
                    "origin_time": [0],
                    "ancestor_list": ["[none]"],
                    "name": "genesis",
                    "taxon_label": "genesis",
                }
            ),
        ],
        ignore_index=True,
        join="inner",
    )

    alifestd_df.reset_index(drop=True, inplace=True)

    return alifestd_df



from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ...juxtaposition import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
    calc_probability_differentia_collision_between,
)


def calc_rank_of_mrca_bounds_provided_confidence_level(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
    prior: str,
    requested_confidence_level: float = 0.95,
) -> float:
    """Calculate provided confidence for a MRCA generation estimate.

    With what actual confidence is the true rank of the MRCA captured within the calculated estimate bounds for a requested confidence level?
    Guaranteed greater than or equal to the requested confidence level.

    The same argument may be provided for focal and other.
    """
    if prior != "arbitrary":
        raise NotImplementedError
    assert 0.0 <= requested_confidence_level <= 1.0
    if other is not None:
        assert (
            focal.GetStratumDifferentiaBitWidth()
            == other.GetStratumDifferentiaBitWidth()
        )

    n = calc_min_implausible_spurious_consecutive_differentia_collisions_between(
        focal,
        other,
        significance_level=1 - requested_confidence_level,
    )
    res = 1 - calc_probability_differentia_collision_between(focal, other) ** n
    assert res >= requested_confidence_level
    assert 0 <= res <= 1
    return res



import numpy as np

from hstrat._auxiliary_lib import numpy_index


def test_numpy_index_first_occurrence():
    # Test that the function returns the correct index for the first occurrence of an item
    a = np.array([1, 2, 3, 4, 5])
    assert numpy_index(a, 3) == (2,)
    assert a[numpy_index(a, 3)] == 3
    assert numpy_index(a, 5) == (4,)
    assert a[numpy_index(a, 5)] == 5
    assert numpy_index(a, 1) == (0,)
    assert a[numpy_index(a, 1)] == 1

    b = np.array([0, 1, 2, 1, 4])
    assert numpy_index(b, 1) == (1,)
    assert b[numpy_index(b, 1)] == 1

    c = np.array([1])
    assert numpy_index(c, 1) == (0,)
    assert c[numpy_index(c, 1)] == 1


def test_numpy_index_reverse_index():
    # Test that the function returns the correct index for a reverse indexed # array
    a = np.array([5, 4, 3, 2, 1])
    assert numpy_index(a[::-1], 3) == (2,)
    assert a[::-1][numpy_index(a[::-1], 3)] == 3

    assert numpy_index(a[::-1], 4) == (3,)
    assert a[::-1][numpy_index(a[::-1], 4)] == 4


def test_numpy_index_item_not_found():
    # Test that the function returns None when the item is not found in the
    # array
    a = np.array([1, 2, 3, 4, 5])
    assert numpy_index(a, 6) is None

    b = np.array([0, 1, 2, 1, 4])
    assert numpy_index(b, 3) is None

    c = np.array([1])
    assert numpy_index(c, 2) is None


def test_numpy_index_empty_array():
    # Test that the function returns None when the array is empty
    a = np.array([])
    assert numpy_index(a, 1) is None



import pytest

from hstrat._auxiliary_lib import generate_n


@pytest.fixture
def mock_generator() -> str:
    count = 0

    def generator():
        nonlocal count
        count += 1
        return f"mock data {count - 1}"

    return generator


@pytest.mark.parametrize("n", range(10))
def test_generate_n(mock_generator, n):
    results = [*generate_n(mock_generator, n)]
    assert results == [f"mock data {i}" for i in range(n)]



import pytest

from hstrat.hstrat import HereditaryStratigraphicColumn, pseudostochastic_algo


@pytest.mark.parametrize(
    "hash_salt",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_policy_consistency(hash_salt):
    policy = pseudostochastic_algo.Policy(hash_salt)
    spec = policy.GetSpec()
    instance = pseudostochastic_algo.CalcNumStrataRetainedUpperBound(spec)
    column = HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
    )
    for num_strata_deposited in range(1, 10**3):
        policy_requirement = column.GetNumStrataRetained()
        for which in (
            instance,
            pseudostochastic_algo.CalcNumStrataRetainedUpperBound(spec),
        ):
            assert (
                which(
                    policy,
                    num_strata_deposited,
                )
                >= policy_requirement
            )


@pytest.mark.parametrize(
    "hash_salt",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_eq(hash_salt):
    policy = pseudostochastic_algo.Policy(hash_salt)
    spec = policy.GetSpec()
    instance = pseudostochastic_algo.CalcNumStrataRetainedUpperBound(spec)

    assert instance == instance
    assert instance == pseudostochastic_algo.CalcNumStrataRetainedUpperBound(
        spec,
    )
    assert instance is not None



"""Tools to specify stratum retention policy."""

from . import (
    stratum_retention_algorithms,
    stratum_retention_policy_evaluators,
    stratum_retention_policy_parameterizers,
)
from .stratum_retention_algorithms import *  # noqa: F401
from .stratum_retention_policy_evaluators import *  # noqa: F401
from .stratum_retention_policy_parameterizers import *  # noqa: F401

__all__ = (
    stratum_retention_algorithms.__all__
    + stratum_retention_policy_evaluators.__all__
    + stratum_retention_policy_parameterizers.__all__
)



def capitalize_n(string: str, n: int) -> str:
    """Create a copy of `string` with the first `n` characers capitalized.

    If `n` is negative, the last `n` characters will be capitalized.

    Examples
    --------
    >>> capitalize_n('hello world', 2)
    'HEllo world'
    >>> capitalize_n('goodbye', 4)
    'GOODbye'
    """
    return (
        string[:n].upper() + string[n:]
        if n >= 0
        else string[:n] + string[n:].upper()
    )



"""Classes to specify prior expectation for the distribution of MRCA generation
between hereditary stratigraphic columns, to assist in estimating the
phylogenetic relationship of hereditary stratigraphic columns."""

from ._ArbitraryPrior import ArbitraryPrior
from ._ExponentialPrior import ExponentialPrior
from ._GeometricPrior import GeometricPrior
from ._UniformPrior import UniformPrior

__all__ = [
    "ArbitraryPrior",
    "GeometricPrior",
    "ExponentialPrior",
    "UniformPrior",
]

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import itertools as it
import unittest

from hstrat._auxiliary_lib import memoize_generator


class TestMemoizeGenerator(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):
        @memoize_generator()
        def empty_generator(*args, **kwargs):
            return
            yield

        assert [*empty_generator()] == []
        assert [*empty_generator()] == []
        assert [*empty_generator(1)] == []
        assert [*empty_generator(2)] == []

    def test_singleton(self):
        @memoize_generator()
        def singleton_generator(item):
            yield item

        assert [*singleton_generator("a")] == ["a"]
        assert [*singleton_generator("b")] == ["b"]
        assert [*singleton_generator("a")] == ["a"]
        assert [*singleton_generator("a")] == ["a"]

    def test_infinite(self):
        @memoize_generator()
        def infinite_generator(i):
            for x in it.count(i):
                yield x

        assert [*it.islice(infinite_generator(0), 3)] == [*range(3)]
        assert [*it.islice(infinite_generator(0), 2)] == [*range(2)]
        assert [*it.islice(infinite_generator(1), 2)] == [*range(1, 3)]


if __name__ == "__main__":
    unittest.main()



import numpy as np


class ArbitraryPrior:
    """Enacts a prior probability density distribution on the generation of the
    most recent common ancestor (MRCA) between extant hereditary stratigraphic
    columns that is arbitrary, but computationally efficient.

    The prior expectation for MRCA generation is taken as equal probability
    within each interval between ranks with common strata retained by both
    extant columns up through the first retained disparity bewteen the columns.

    Prior probability density is assumed uniformly distributed within each
    interval between coincident retained ranks. So, conditioning on the
    assumption that the true generation of the MRCA occurs within a particular
    interval, the prior expected value for the MRCA generation will be the
    midpoint of the interval.

    This prior is simple to compute, but may not meaningfully reflect the
    a reasonable pre-expectation for the MRCA generation. Importantly, the
    enacted prior expectation will depend directly on the instrumentation used
    (i.e., the distribution of coincident retained strata induced by the chosen
    stratum retention policy). For example, a wide interval between coincident
    retained ranks and a short interval between coincident retaine dranks will
    be assigned equal prior probability, resulting in greater per-generation
    prior probability within the small window than within a wide window.

    This prior policy guarantees the maximum likelihood estimate to fall
    between the last retained commonality and the first retained disparity of
    two extant columns. Because each interval between coincident retained ranks
    has equal prior probability, the likelihood of the true MRCA falling within
    preceding intervals strictly decreases with qualification by spurious
    differentia collisions (i.e., common retained strata). This property makes
    maximum likelihood estimation under this prior especially efficient.
    """

    def CalcIntervalProbabilityProxy(
        self: "ArbitraryPrior", begin_rank: int, end_rank: int
    ) -> float:
        """Characterize the prior probability of the MRCA generation falling
        within an interval range.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        float
            The proxy statistic, proportional to the true estimated interval
            probability of the MRCA value by a fixed (but unspecified) constant
            proportion.
        """
        return 1.0

    def CalcIntervalConditionedMean(
        self: "ArbitraryPrior", begin_rank: int, end_rank: int
    ) -> float:
        """Calcualate the centriod of prior probability mass within an interval
        of possible MRCA generations.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        float
            The prior expected generation of MRCA conditioned on the assumption
            that the MRCA falls within the given interval.
        """
        return np.mean((begin_rank, end_rank - 1.0))



import sys
import typing
import warnings

import opytional as opyt

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ...juxtaposition._impl import dispatch_impl
from ._calc_rank_of_earliest_detectable_mrca_between import (
    calc_rank_of_earliest_detectable_mrca_between,
)
from ._does_have_any_common_ancestor import does_have_any_common_ancestor


def calc_rank_of_mrca_bounds_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    prior: str,
    confidence_level: float = 0.95,
    strict=True,
) -> typing.Optional[typing.Tuple[int, int]]:
    """Within what generation range did MRCA fall?

    Calculate bounds on estimate for the number of depositions elapsed
    along the line of descent before the most recent common ancestor with
    second.

    Parameters
    ----------
    prior : {"arbitrary"}
        Prior probability density distribution over possible generations of the
        MRCA.

        Currently only "arbitrary" supported.
    confidence_level : float, optional
        Bounds must capture what probability of containing the true rank of
        the MRCA? Default 0.95.

    Returns
    -------
    (int, int), optional
        Inclusive lower and then exclusive upper bound on estimate or None
        if no common ancestor between first and second can be resolved with
        sufficient confidence. (Sufficient confidence depends on
        confidence_level.)

    See Also
    --------
    calc_rank_of_mrca_uncertainty_between :
        Wrapper to report uncertainty of calculated bounds.
    calc_rank_of_earliest_detectable_mrca_between :
        Could any MRCA be detected between first and second? What is the rank
        of the earliest MRCA that could be reliably detected?
    calc_rank_of_mrca_bounds_provided_confidence_level :
        With what actual confidence (i.e., more than requested) is the true
        rank of the MRCA captured within the calculated bounds?
    does_definitively_have_no_common_anestor :
        Does the hereditary stratigraphic record definitively prove that first
        and second could not possibly share a common ancestor?

    Notes
    -----
    The true rank of the MRCA is guaranteed to never fall above the returned
    bounds but may fall below.

    An alternate approach could be to construct the bounds such that the
    true rank of the MRCA will fall above or below the bounds with equal
    probability. This would involve setting the confidence level for
    calculating the first disparity with second to significance_level/2 and
    the confidence level for calculaing the last comonality with second to
    1 - significance_level/2. This means the confidence level applied to
    calculating the first disparity with second would always be <= 0.5.
    However, shifting the calculated first disparity with second below
    the definitive max first retained disparity requires confidence level
    >= 0.5. So, in practice such a symmetric approach would only result in
    the lower bound being shifted downward. For this reason, it is no longer
    provided as an option.

    In the absence of evidence to the contrary (i.e., more common
    strata than spurious differentia collisions alone could plausibly
    cause), this method assumes no common ancestry between first and second,
    returning None. This means that if few enough common ranks are shared
    between first and second (and the differentia bit with is small enough),
    it may not be possible to detect any common ancestry after accounting
    for the possibility of spurious differentia collisions (even if common
    ancestry did exist). So, calls to this method would always return None.
    Likewise, MRCAs at very early ranks may not be able to be reliably
    detected due to insufficient evidence. This can lead to cases where
    columns with true common ancestry have MRCA bounds estimated as None at
    much higher than the expected failure rate at the given confidence
    level. Note that with sufficient differentia bit width (i.e., so that
    even one collision is implausible at the given confidence level) this
    issue does not occur. Use CalcRankOfEarliestDetectableMrcaWith to
    determine the earliest rank at which an MRCA could be reliably detected
    between first and second.
    """
    if prior != "arbitrary":
        raise NotImplementedError
    assert 0.0 <= confidence_level <= 1.0

    if (
        not sys.flags.optimize
        and calc_rank_of_earliest_detectable_mrca_between(
            first,
            second,
            confidence_level=confidence_level,
        )
        is None
    ):
        warnings.warn(
            "Insufficient common ranks between columns to detect common "
            "ancestry at given confidence level."
        )

    if (
        does_have_any_common_ancestor(
            first,
            second,
            confidence_level=confidence_level,
        )
        or not strict
    ):
        last_commonality, first_disparity = dispatch_impl(
            first, second
        ).calc_rank_of_parity_segue_between(
            first,
            second,
            confidence_level_commonality=confidence_level,
            confidence_level_disparity=0.49,  # 0.49 i.e., definitive max
        )
        assert (
            first_disparity is not None
            or first.GetNumStrataDeposited() == second.GetNumStrataDeposited()
        )
        return (
            opyt.or_value(last_commonality, 0),
            opyt.or_value(first_disparity, first.GetNumStrataDeposited()),
        )
    else:
        return None



import typing

from ...._auxiliary_lib import anytree_iterative_deepcopy
from .._impl import TrieInnerNode


class CompoundTriePostprocessor:
    """Functor to sequentially apply multiple trie postprocessors.

    Allows for the combination and sequential application of multiple trie
    postprocessors.
    """

    _postprocessors: typing.Iterable[typing.Callable]

    def __init__(
        self: "CompoundTriePostprocessor",
        postprocessors: typing.Iterable[typing.Callable],
    ) -> None:
        """Initialize functor instance.

        Parameters
        ----------
        postprocessors : typing.Iterable[typing.Callable]
            The sequence of postprocess functors to be applied.
        """
        self._postprocessors = postprocessors

    def __call__(
        self: "CompoundTriePostprocessor",
        trie: TrieInnerNode,
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> TrieInnerNode:
        """Apply stored postprocessors in sequence.

        Parameters
        ----------
        trie : TrieInnerNode
            The input trie to be postprocessed.
        p_differentia_collision : float
            Probability of a randomly-generated differentia matching an
            existing differentia.

            Forwarded to the postprocess functors.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode
            The postprocessed trie after applying all stored postprocessors.
        """
        if not mutate:
            trie = anytree_iterative_deepcopy(
                trie, progress_wrap=progress_wrap
            )

        for postprocessor in self._postprocessors:
            trie = postprocessor(
                trie=trie,
                p_differentia_collision=p_differentia_collision,
                mutate=True,
                progress_wrap=progress_wrap,
            )

        return trie



import copy
import random
import typing

import anytree
import opytional as opyt

from ...._auxiliary_lib import generate_n, render_to_base64url
from ._TrieLeafNode import TrieLeafNode


class TrieInnerNode(anytree.NodeMixin):
    """Inner node of a trie (a.k.a. prefix tree) of differentia
    sequences of hereditary stratigraphic artifacts within a population.
    Each inner node represents the hypothesized origination of a particular
    allele (i.e., differentia at a particular rank).

    Note that more than one` TrieInnerNode` representing a particular
    rank/differentia combination may be present. This is only the case when the
    colliding alleles are known to have independent origins due to previous
    disparities in the phylogenetic record.

    Only `TrieLeafNode` instances will occupy leaf node positions in the trie
    structure. Each `TrieLeafNode` represents a single member of the extant
    artifact population. An instance of `TrieInnerNode` may have child nodes of
    both types.
    """

    _rank: int  # rank of represented allele
    _differentia: int  # differentia of represented allele
    _tiebreaker: int  # random 128-bit integer used to break ties.

    def __init__(
        self: "TrieInnerNode",
        rank: typing.Optional[int] = None,
        differentia: typing.Optional[int] = None,
        parent: typing.Optional["TrieInnerNode"] = None,
    ) -> None:
        """Initialize a `TrieInnerNode` instance.

        Parameters
        ----------
        rank : int, optional
            The rank of the node, or None for shim "ancestor of all geneses" root node.

            Pass None for root node, int otherwise
        differentia : int, optional
            The fingerprint or differentia of the node, or None for shim
            "ancestor of all geneses" root node.
        parent : int, optional
            The parent node, or None for shim "ancestor of all geneses" root
            node.
        """
        self.parent = parent
        self._rank = rank
        self._differentia = differentia
        assert (self._rank is None) == (self._differentia is None)
        self._tiebreaker = random.getrandbits(128)  # uuid standard 128 bits

    def IsAnOriginationOfAllele(
        self: "TrieInnerNode",
        rank: int,
        differentia: int,
    ) -> bool:
        """Checks if the node represents an origination of a given allele."""
        return self._rank == rank and self._differentia == differentia

    def FindOriginationsOfAllele(
        self: "TrieInnerNode",
        rank: int,
        differentia: int,
    ) -> typing.Iterator["TrieInnerNode"]:
        """Searches subtree to find all possible `TrieInnerNodes` representing
        an origination of the specified `rank`-`differentia` allele."""
        assert rank is not None
        assert differentia is not None
        # note: assert handles root case
        assert (self._differentia is None) == (self._rank is None)

        search_stack = [self]
        while search_stack:
            current_node = search_stack.pop()

            # handle root node case
            if current_node._rank is None:
                search_stack.extend(current_node.inner_children)
            # If the current node matches the rank and differentia, yield it
            elif current_node.IsAnOriginationOfAllele(rank, differentia):
                yield current_node
            # If the target allele's rank is deeper than the current node,
            # keep searching
            elif rank > current_node._rank:
                search_stack.extend(current_node.inner_children)

    def GetDeepestCongruousAlleleOrigination(
        self: "TrieInnerNode",
        taxon_allele_iter: typing.Iterator[typing.Tuple[int, int]],
    ) -> (
        typing.Tuple["TrieInnerNode", typing.Iterator[typing.Tuple[int, int]]]
    ):
        """Descends the subtrie to retrieve the deepest prefix consistent with
        the hereditary stratigraphic record of a query taxon.

        In the case where more than one possible originations are found for an
        allele, all will be searched recursively for further matches with the
        next alleles in the focal taxon's hereditary stratigraphic record.
        (This scenario occurs when the focal taxon has discarded a strata that
        differentiates subsequently colliding alleles among earlier taxa).
        Nodes' `_tiebreaker` fields determinstically resolve any ambiguities
        for the deepest consistent prefix that may aries in such cases.

        Parameters
        ----------
        taxon_allele_iter : typing.Iterator[typing.Tuple[int, int]]
            An iterator over the taxon alleles (rank/differentia pairs) in rank-
            ascending order.

            Must be copyable.

        Returns
        -------
        typing.Tuple["TrieInnerNode", typing.Iterator[typing.Tuple[int, int]]]
            A tuple containing the retrieved deepest prefix tree match and an
            iterator over the taxon alleles remaining past the retrieved allele.
        """
        # iterator must be copyable
        assert [*copy.copy(taxon_allele_iter)] == [
            *copy.copy(taxon_allele_iter)
        ]

        node_stack = [self]
        allele_origination_stack = [taxon_allele_iter]
        deepest_origination = self
        deepest_taxon_allele_iter = copy.copy(taxon_allele_iter)
        while True:

            candidate_origination = node_stack.pop()
            taxon_allele_iter = allele_origination_stack.pop()

            # Update the deepest origination if the candidate is deeper
            if (
                opyt.or_value(candidate_origination._rank, -1),
                candidate_origination._tiebreaker,
            ) > (
                opyt.or_value(deepest_origination._rank, -1),
                deepest_origination._tiebreaker,
            ):
                deepest_origination = candidate_origination
                deepest_taxon_allele_iter = copy.copy(taxon_allele_iter)

            # If taxon has susbsequent allele, search its origination
            next_allele = next(taxon_allele_iter, None)
            if next_allele is not None:
                node_stack.extend(
                    candidate_origination.FindOriginationsOfAllele(
                        *next_allele
                    )
                )
                allele_origination_stack.extend(
                    generate_n(
                        lambda: copy.copy(taxon_allele_iter),
                        len(node_stack) - len(allele_origination_stack),
                    )
                )

            assert len(node_stack) == len(allele_origination_stack)

            if not node_stack:  # no more nodes to explore
                break

        return deepest_origination, deepest_taxon_allele_iter

    def InsertTaxon(
        self: "TrieInnerNode",
        taxon_label: str,
        taxon_allele_iter: typing.Iterator[typing.Tuple[int, int]],
    ) -> TrieLeafNode:
        """Inserts a taxon into the trie, ultimately resulting in the creation
        of an additional `TrieLeafNode` leaf and --- if necessary --- a
        unifurcating chain of `TrieInnerNode`'s subtending it.

        Parameters
        ----------
        taxon_label : str
            The label of the taxon to be inserted.
        taxon_allele_iter : typing.Iterator[typing.Tuple[int, int]]
            An iterator over the taxon's remaining allele sequence that has not
            yet been accounted for thus deep into the trie.

        Notes
        -----
        Should only be called on the node corresponding to the deepest congrous
        allele origination found via `GetDeepestCongruousAlleleOrigination`.
        However, may be called on the trie root node when the entire artifact
        population has identical deposition counts. (Because, given a
        deterministic stratum retention strategy, insertion is greatly
        simplified because no colliding allele originations can arise due to
        the impossibility of preceding divergences not retained by the taxon
        being inserted.)

        Returns
        -------
        TrieLeafNode
            The leaf node representing the inserted taxon.
        """
        cur_node = self
        for next_rank, next_differentia in taxon_allele_iter:
            assert next_rank is not None
            assert next_differentia is not None
            for child in cur_node.inner_children:
                # check immediate children for next allele
                #
                # common allele origination trace is for special condition
                # optimization where GetDeepestCongruousAlleleOrigination
                # isn't needed
                if child.IsAnOriginationOfAllele(next_rank, next_differentia):
                    cur_node = child
                    break
            else:
                # if no congruent node exists, create a new TrieInnerNode
                cur_node = TrieInnerNode(
                    next_rank, next_differentia, parent=cur_node
                )

        # create a TrieLeafNode representing the inserted taxon
        return TrieLeafNode(parent=cur_node, taxon_label=taxon_label)

    @property
    def taxon_label(self: "TrieInnerNode") -> str:
        """Programatically-generated unique identifier for internal node,
        intended to translate into a unique label for internal taxa after
        conversion to a phylogenetic reconstruction."""
        if self.parent is None:
            return "Root"
        else:
            # numpy ints cause indexing errors; convert to native int
            # uid field necessary to distinguish colliding allele originations
            return f"""Inner+r={self._rank}+d={
                render_to_base64url(int(self._differentia))
            }+uid={
                render_to_base64url(int(self._tiebreaker))
            }"""

    @property
    def taxon(self: "TrieInnerNode") -> str:
        """Alias for taxon_label."""
        return self.taxon_label

    @property
    def rank(self: "TrieInnerNode") -> int:
        return opyt.or_value(self._rank, 0)

    @property
    def inner_children(
        self: "TrieInnerNode",
    ) -> typing.Iterator["TrieInnerNode"]:
        """Returns iterator over non-leaf child nodes."""
        return filter(
            lambda child_: isinstance(child_, TrieInnerNode),
            self.children,
        )

    @property
    def outer_children(
        self: "TrieInnerNode",
    ) -> typing.Iterator["TrieLeafNode"]:
        """Returns iterator over leaf child nodes."""
        return filter(
            lambda child_: isinstance(child_, TrieLeafNode),
            self.children,
        )

    def __repr__(self: "TrieInnerNode") -> str:
        return f"""(rank {
            self._rank
        }, diff {
            self._differentia
        }) @ {
            render_to_base64url(id(self) % 8192)
        }"""



"""Strata storage implementations for use with HereditaryStratigraphicColumn."""

from ._HereditaryStratumOrderedStoreDict import (
    HereditaryStratumOrderedStoreDict,
)
from ._HereditaryStratumOrderedStoreList import (
    HereditaryStratumOrderedStoreList,
)
from ._HereditaryStratumOrderedStoreTree import (
    HereditaryStratumOrderedStoreTree,
)

provided_stratum_ordered_stores = [
    HereditaryStratumOrderedStoreDict,
    HereditaryStratumOrderedStoreList,
    HereditaryStratumOrderedStoreTree,
]

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "HereditaryStratumOrderedStoreDict",
    "HereditaryStratumOrderedStoreList",
    "HereditaryStratumOrderedStoreTree",
    "provided_stratum_ordered_stores",
]

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import numpy as np
import pytest

from hstrat.hstrat import depth_proportional_resolution_tapered_algo


@pytest.mark.parametrize(
    "depth_proportional_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10**2,
        ),
        (2**32,),
    ],
)
def test_policy_consistency(depth_proportional_resolution, time_sequence):
    policy = depth_proportional_resolution_tapered_algo.Policy(
        depth_proportional_resolution
    )
    spec = policy.GetSpec()
    instance = depth_proportional_resolution_tapered_algo.CalcMrcaUncertaintyAbsUpperBound(
        spec,
    )
    for num_strata_deposited in time_sequence:
        for actual_mrca_rank in (
            np.random.default_rng(num_strata_deposited,).integers(
                low=0,
                high=num_strata_deposited,
                size=10**2,
            )
            if num_strata_deposited
            else iter(())
        ):
            policy_requirement = policy.CalcMrcaUncertaintyAbsExact(
                num_strata_deposited,
                num_strata_deposited,
                actual_mrca_rank,
            )
            for which in (
                instance,
                depth_proportional_resolution_tapered_algo.CalcMrcaUncertaintyAbsUpperBound(
                    spec
                ),
            ):
                assert (
                    which(
                        policy,
                        num_strata_deposited,
                        num_strata_deposited,
                        actual_mrca_rank,
                    )
                    >= policy_requirement
                )


@pytest.mark.parametrize(
    "depth_proportional_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_eq(depth_proportional_resolution):
    policy = depth_proportional_resolution_tapered_algo.Policy(
        depth_proportional_resolution
    )
    spec = policy.GetSpec()
    instance = depth_proportional_resolution_tapered_algo.CalcMrcaUncertaintyAbsUpperBound(
        spec
    )

    assert instance == instance
    assert (
        instance
        == depth_proportional_resolution_tapered_algo.CalcMrcaUncertaintyAbsUpperBound(
            spec,
        )
    )
    assert instance is not None


@pytest.mark.parametrize(
    "depth_proportional_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_negative_index(depth_proportional_resolution):
    policy = depth_proportional_resolution_tapered_algo.Policy(
        depth_proportional_resolution
    )
    spec = policy.GetSpec()
    instance = depth_proportional_resolution_tapered_algo.CalcMrcaUncertaintyAbsUpperBound(
        spec
    )

    for diff in range(1, 100):
        assert instance(policy, 100, 100, -diff,) == instance(
            policy,
            100,
            100,
            99 - diff,
        )

        assert instance(policy, 101, 100, -diff,) == instance(
            policy,
            101,
            100,
            99 - diff,
        )

        assert instance(policy, 150, 100, -diff,) == instance(
            policy,
            150,
            100,
            99 - diff,
        )

        assert instance(policy, 100, 101, -diff,) == instance(
            policy,
            101,
            100,
            99 - diff,
        )

        assert instance(policy, 100, 150, -diff,) == instance(
            policy,
            150,
            100,
            99 - diff,
        )



import alifedata_phyloinformatics_convert as apc
import tqdist


def tree_quartet_distance(x, y) -> float:
    tree_a = apc.RosettaTree(x).as_dendropy
    tree_b = apc.RosettaTree(y).as_dendropy

    # must suppress root unifurcations or tqdist barfs
    # see https://github.com/uym2/tripVote/issues/15
    tree_a.unassign_taxa(exclude_leaves=True)
    tree_a.suppress_unifurcations()
    tree_b.unassign_taxa(exclude_leaves=True)
    tree_b.suppress_unifurcations()

    tree_a_taxon_labels = [
        leaf.taxon.label for leaf in tree_a.leaf_node_iter()
    ]
    tree_b_taxon_labels = [
        leaf.taxon.label for leaf in tree_b.leaf_node_iter()
    ]
    assert {*tree_a_taxon_labels} == {*tree_b_taxon_labels}
    for taxon_label in tree_a_taxon_labels:
        assert taxon_label
        assert taxon_label.strip()

    return tqdist.quartet_distance(
        tree_a.as_string(schema="newick").strip(),
        tree_b.as_string(schema="newick").strip(),
    )



import typing

from ...._detail import PolicyCouplerBase
from ...._impl import GenDropRanksFromPredKeepRank
from ..._PolicySpec import PolicySpec
from ..._impl import get_retained_ranks


class _PredKeepRank:
    """Functor to implement the approximate space-filling MRCA-recency-
    proportional resolution stratum retention policy, for use with
    HereditaryStratigraphicColumn.

    This functor enacts the approximate space-filling MRCA-recency-
    proportional resolution stratum retention policy by specifying the set of
    strata ranks that should be purged from a hereditary stratigraphic column
    when the nth stratum is deposited.
    """

    def __init__(
        self: "_PredKeepRank",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "_PredKeepRank", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    @staticmethod
    def _do_call(
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        stratum_rank: int,
    ) -> bool:
        """Implementation for __call__ to faciliate external (but within-
        library) calls."""
        spec = policy.GetSpec()
        degree = spec.GetDegree()
        interspersal = spec.GetInterspersal()

        return stratum_rank in get_retained_ranks(
            degree,
            interspersal,
            # +1 because of in-progress deposition
            num_stratum_depositions_completed + 1,
        )

    def __call__(
        self: "_PredKeepRank",
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        stratum_rank: int,
    ) -> bool:
        """Decide if a stratum within the stratagraphic column should be
        retained or purged.

        Every time a new stratum is deposited, this method is called on each
        stratum present in a HereditaryStratigraphicColumn to determine whether
        it should be retained. Strata that return False are immediately purged
        from the column, meaning that for a stratum to persist it must earn a
        True result from this method each and every time a new stratum is
        deposited.

        Parameters
        ----------
        policy: Policy
            Policy this functor enacts.
        stratum_rank : int
            The number of strata that were deposited before the stratum under
            consideration for retention.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.

        Returns
        -------
        bool
            True if the stratum should be retained, False otherwise.

        See Also
        --------
        depth_proportional_resolution_tapered_algo:
            For details on the rationale, implementation, and guarantees of the
            depth-proportional resolution stratum retention policy.
        """
        return _PredKeepRank._do_call(
            policy,
            num_stratum_depositions_completed,
            stratum_rank,
        )


FromPredKeepRank = GenDropRanksFromPredKeepRank(_PredKeepRank)



"""Tools to load and save hereditary stratigraphic columns."""

from ._assemblage_from_records import assemblage_from_records
from ._col_from_packet import col_from_packet
from ._col_from_packet_buffer import col_from_packet_buffer
from ._col_from_records import col_from_records
from ._col_to_dataframe import col_to_dataframe
from ._col_to_packet import col_to_packet
from ._col_to_records import col_to_records
from ._col_to_specimen import col_to_specimen
from ._pack_differentiae import pack_differentiae
from ._pack_differentiae_bytes import pack_differentiae_bytes
from ._pack_differentiae_str import pack_differentiae_str
from ._policy_from_records import policy_from_records
from ._policy_to_records import policy_to_records
from ._pop_from_records import pop_from_records
from ._pop_to_assemblage import pop_to_assemblage
from ._pop_to_dataframe import pop_to_dataframe
from ._pop_to_records import pop_to_records
from ._specimen_from_records import specimen_from_records
from ._unassemblage_from_records import unassemblage_from_records
from ._unpack_differentiae import unpack_differentiae
from ._unpack_differentiae_bytes import unpack_differentiae_bytes
from ._unpack_differentiae_str import unpack_differentiae_str

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "assemblage_from_records",
    "col_to_dataframe",
    "col_from_packet",
    "col_from_packet_buffer",
    "col_from_records",
    "col_to_packet",
    "col_to_records",
    "col_to_specimen",
    "pack_differentiae",
    "pack_differentiae_bytes",
    "pack_differentiae_str",
    "policy_from_records",
    "policy_to_records",
    "pop_from_records",
    "pop_to_assemblage",
    "pop_to_dataframe",
    "pop_to_records",
    "specimen_from_records",
    "unassemblage_from_records",
    "unpack_differentiae",
    "unpack_differentiae_bytes",
    "unpack_differentiae_str",
]

from .._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import typing


def give_len(wrapee: typing.Iterable, len_: int) -> typing.Iterable:
    """Wrap an object to make Sized with `len(wrapee)` as provided `len_`."""

    # adapted from https://stackoverflow.com/a/2982
    class GiveLenWrapper:
        _wrapee: typing.Iterable

        def __init__(self: "GiveLenWrapper", wrapee: typing.Iterable) -> None:
            self._wrapee = wrapee

        # must define members inline or won't be recognized as wrapee etc.
        for attr in dir(wrapee):
            if attr not in {
                "__class__",
                "__delattr__",
                "__getattribute__",
                "__init__",
                "__name__",
                "__new__",
                "__qualname__",
                "__setattr__",
            }:
                exec(
                    f"""
{'@property' if not callable(getattr(wrapee, attr)) else ''}
def {attr}(self, *args, **kwargs):
    return self._wrapee.{attr}(*args, **kwargs)
                    """
                )

        def __len__(self: "GiveLenWrapper") -> int:
            return len_

    return GiveLenWrapper(wrapee)



from hstrat._auxiliary_lib import render_to_base64url


def test_render_to_base64url():
    assert render_to_base64url(163) == "Cj"
    assert render_to_base64url(63) == "_"
    assert render_to_base64url(62) == "-"
    assert render_to_base64url(42) == "q"
    assert render_to_base64url(7) == "H"
    assert render_to_base64url(0) == "A"



import unittest

from hstrat._auxiliary_lib import omit_last


class TestOmitLast(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test(self):

        for len_ in range(10):
            operand = [*range(len_)]
            assert [*omit_last(operand)] == operand[:-1]
            assert [*omit_last(iter(operand))] == operand[:-1]


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd

from hstrat._auxiliary_lib import alifestd_find_root_ids


def test_alifestd_find_root_ids_returns_expected_output_with_single_root():
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "ancestor_list": ["[none]", "1", "2"],
        }
    )
    expected_output = np.array([1])
    for df in phylogeny_df, phylogeny_df.sample(frac=1):
        output = alifestd_find_root_ids(df)
        assert np.array_equal(output, expected_output)


def test_alifestd_find_root_ids_returns_expected_output_with_multiple_roots():
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "ancestor_list": ["[none]", "[none]", "1", "2"],
        }
    )
    expected_output = np.array([1, 2])
    for df in phylogeny_df, phylogeny_df.sample(frac=1):
        output = alifestd_find_root_ids(df)
        assert np.array_equal(output, expected_output)


def test_alifestd_find_root_ids_does_not_mutate_input_dataframe():
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "ancestor_list": ["[none]", "1", "2"],
        }
    )
    input_df = phylogeny_df.copy()
    alifestd_find_root_ids(input_df)
    assert phylogeny_df.equals(input_df)



import typing

from .._auxiliary_lib import HereditaryStratigraphicArtifact
from ._impl import dispatch_impl


def calc_rank_of_first_retained_disparity_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    confidence_level: float = 0.95,
) -> typing.Optional[int]:
    """Determine upper bound on MRCA generation at given confidence.

    How many depositions elapsed along the columns' lines of
    descent before the first mismatching strata at the same rank between
    first and second?

    Parameters
    ----------
    confidence_level : float, optional
        With what probability should the true rank of the first disparity
        between first and second fall at or after the returned rank? Default
        0.95.

    Returns
    -------
    int, optional
        The number of depositions elapsed or None if no disparity (i.e.,
        both columns have same number of strata deposited and the most
        recent stratum is common between first and second).

    Notes
    -----
    If no mismatching strata are found but first and second have different
    numbers of strata deposited, this method returns one greater than the
    lesser of the columns' deposition counts.

    The true rank of the first disparity with second is guaranteed to never
    be after the returned rank when confidence_level < 0.5.

    If the differentia width and confidence level are configured such that
    one or more suprious differentia collisions is plausible, this method
    will never return None.

    Consider two columns that compare identical at all common ranks. If
    these columns have few enough common ranks that all these collisions
    could plausibly be spurious then there is not yet enough to support
    common lineage between the columns so rank 0 must be selected as the
    estimated site of first retained disparity. If these columns have
    enough common ranks that all collisions could not plausibly be spurious
    then the largest rank that could not conceivably be spurious must be
    selected (and we are guaranteed such a rank exists). It is only when
    not even one collision could plausibly be spurious that strong enough
    evidence exists to conclude that there is no disparity.

    For example, the method may return None with 64-bit differentia and 95%
    confidence level but will never return None with 1-bit differentia and
    95% confidence level.
    """
    assert 0.0 <= confidence_level <= 1.0
    return dispatch_impl(
        first, second
    ).calc_rank_of_first_retained_disparity_between(
        first, second, confidence_level
    )



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_make_ancestor_id_col,
    alifestd_make_ancestor_list_col,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_make_ancestor_list_col(phylogeny_df):

    phylogeny_df["ancestor_id"] = alifestd_make_ancestor_id_col(
        phylogeny_df["id"], phylogeny_df["ancestor_list"]
    )
    assert all(
        alifestd_make_ancestor_list_col(
            phylogeny_df["id"], phylogeny_df["ancestor_id"]
        ).str.lower()
        == phylogeny_df["ancestor_list"].str.lower()
    )


def test_alifestd_make_ancestor_list_col_empty():
    res = alifestd_make_ancestor_list_col(
        pd.Series(dtype=int), pd.Series(dtype=int)
    )
    assert len(res) == 0
    assert pd.api.types.is_string_dtype(res.dtype)


def test_alifestd_make_ancestor_list_col_root_ancestor_token():
    res = alifestd_make_ancestor_list_col(
        pd.Series([0]),
        pd.Series([0]),
        root_ancestor_token="boop",
    )
    assert len(res) == 1
    pd.testing.assert_series_equal(
        res,
        pd.Series("[boop]"),
    )



import pandas as pd


def alifestd_make_ancestor_list_col(
    ids: pd.Series, ancestor_ids: pd.Series, root_ancestor_token: str = "none"
) -> pd.Series:
    """Translate a column of integer ancestor id values into alife standard
    `ancestor_list` representation.

    The option `root_ancestor_token` will be sandwiched in brackets to create
    the ancestor list entry for genesis organisms. For example, the token
    "None" will yield the entry "[None]" and the token "" will yield the entry
    "[]". Default "none".
    """

    res = ancestor_ids.map("[{}]".format).astype(str)  # specify for empty
    res[ids == ancestor_ids] = f"[{root_ancestor_token}]"

    return res



import sys
import typing

import more_itertools as mit


# adapted from https://docs.python.org/3/library/itertools.html
def zip_strict(*iterables: typing.Iterable) -> typing.Iterator[typing.Tuple]:
    """Polyfill for strict option on zip added in Python 3.10."""
    if sys.version_info >= (3, 10):
        yield from zip(*iterables, strict=True)
    elif len(iterables):
        try:
            yield from mit.zip_equal(*iterables)
        except mit.UnequalIterablesError:
            raise ValueError
    else:
        yield from zip()



import numpy as np

from hstrat._auxiliary_lib import numpy_fromiter_polyfill


def test_numpy_fromiter_polyfill_int32():
    iterable = [1, 2, 3, 4, 5]
    dtype = np.int32
    expected = np.array(iterable, dtype=dtype)
    result = numpy_fromiter_polyfill(iterable, dtype=dtype)
    assert np.array_equal(expected, result)


def test_numpy_fromiter_polyfill_int32_iterator():
    iterable = [1, 2, 3, 4, 5]
    dtype = np.int32
    expected = np.array(iterable, dtype=dtype)
    result = numpy_fromiter_polyfill(iter(iterable), dtype=dtype)
    assert np.array_equal(expected, result)


def test_numpy_fromiter_polyfill_float64():
    iterable = [1.0, 2.0, 3.0, 4.0, 5.0]
    dtype = np.float64
    expected = np.array(iterable, dtype=dtype)
    result = numpy_fromiter_polyfill(iterable, dtype=dtype)
    assert np.array_equal(expected, result)


def test_numpy_fromiter_polyfill_string():
    iterable = ["a", "b", "c", "d", "e"]
    dtype = np.object_
    expected = np.array(iterable, dtype=dtype)
    result = numpy_fromiter_polyfill(iterable, dtype=dtype)
    assert np.array_equal(expected, result)


def test_numpy_fromiter_polyfill_object():
    iterable = [None, "a", None, "b", "c", None, "d", "e", None]
    dtype = np.object_
    expected = np.array(iterable, dtype=dtype)
    result = numpy_fromiter_polyfill(iterable, dtype=dtype)
    assert np.array_equal(expected, result)


def test_numpy_fromiter_polyfill_object_iterator():
    iterable = [None, "a", None, "b", "c", None, "d", "e", None]
    dtype = np.object_
    expected = np.array(iterable, dtype=dtype)
    result = numpy_fromiter_polyfill(iter(iterable), dtype=dtype)
    assert np.array_equal(expected, result)


def test_numpy_fromiter_polyfill_object_empty():
    iterable = []
    dtype = np.object_
    expected = np.array(iterable, dtype=dtype)
    result = numpy_fromiter_polyfill(iter(iterable), dtype=dtype)
    assert np.array_equal(expected, result)



import pytest

from hstrat.stratum_retention_strategy.stratum_retention_algorithms.recency_proportional_resolution_curbed_algo._impl import (
    calc_geom_seq_nth_root_transition_rank,
    calc_provided_resolution,
)


@pytest.mark.parametrize(
    "size_curb",
    reversed(range(1, 1024)),
)
def test_calc_provided_resolution_consistency(size_curb):
    transition_rank = calc_geom_seq_nth_root_transition_rank(size_curb)

    if transition_rank:
        assert calc_provided_resolution(size_curb, transition_rank - 1) >= 0
    assert calc_provided_resolution(size_curb, transition_rank) < 0



import interval_search as inch
import pytest

import hstrat._auxiliary_lib as auxlib
from hstrat.stratum_retention_strategy.stratum_retention_algorithms.recency_proportional_resolution_curbed_algo._impl import (
    calc_geom_seq_nth_root_transition_rank,
    calc_provided_resolution,
)

from ..impl import iter_backing_policy_transition_ranks


@pytest.mark.parametrize(
    "size_curb",
    reversed(range(1, 1024)),
)
def test_iter_backing_policy_transition_ranks(size_curb):

    *test_ranks, gsnra_rank = [
        0,
        *iter_backing_policy_transition_ranks(size_curb),
    ]

    assert sorted(test_ranks) == test_ranks
    assert len(set(test_ranks)) == len(test_ranks)

    assert gsnra_rank == calc_geom_seq_nth_root_transition_rank(size_curb), (
        gsnra_rank,
        test_ranks[-1],
    )

    for tr1, tr2 in auxlib.pairwise(test_ranks[int(size_curb == 1) : -1]):
        assert tr1 < tr2
        assert tr1 == 0 or calc_provided_resolution(
            size_curb, tr1 - 1
        ) > calc_provided_resolution(size_curb, tr1)
        assert calc_provided_resolution(
            size_curb, tr2 - 1
        ) > calc_provided_resolution(size_curb, tr2)

        assert calc_provided_resolution(
            size_curb, tr1
        ) > calc_provided_resolution(size_curb, tr2)
        assert calc_provided_resolution(
            size_curb, tr1
        ) == calc_provided_resolution(size_curb, tr2 - 1)

    assert (
        len(test_ranks) == 0
        or test_ranks[-1] == 0
        or calc_provided_resolution(size_curb, test_ranks[-1]) >= 0
    )

    assert (
        gsnra_rank == 0
        or calc_provided_resolution(size_curb, gsnra_rank - 1) >= 0
    )
    assert calc_provided_resolution(size_curb, gsnra_rank + 1) < 0
    assert calc_provided_resolution(size_curb, gsnra_rank) < 0


def _iter_backing_policy_transition_ranks_naive(size_curb):
    gsnra_rank = calc_geom_seq_nth_root_transition_rank(size_curb)

    cur_rank = 0
    while cur_rank < gsnra_rank:
        cur_resolution = calc_provided_resolution(size_curb, cur_rank)
        cur_rank = inch.binary_search(
            lambda r: calc_provided_resolution(size_curb, r) != cur_resolution,
            cur_rank,
            gsnra_rank,
        )
        yield cur_rank


@pytest.mark.parametrize(
    "size_curb",
    reversed(range(1, 1024)),
)
def test_iter_backing_policy_transition_ranks_naive(size_curb):
    assert [*iter_backing_policy_transition_ranks(size_curb)] == [
        *_iter_backing_policy_transition_ranks_naive(size_curb)
    ]



import typing

import numpy as np
import ordered_set as ods
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_find_leaf_ids(phylogeny_df: pd.DataFrame) -> typing.List[int]:
    """What ids are not listed in any `ancestor_list`?

    Input dataframe is not mutated by this operation.
    """

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df)
        if "ancestor_id" in phylogeny_df:

            # root is self ref, but must exclude to handle only-root phylo
            internal_node_idxs = phylogeny_df.loc[
                phylogeny_df["ancestor_id"] != phylogeny_df["id"],
                "ancestor_id",
            ].to_numpy()

            leaf_pos_filter = np.ones(len(phylogeny_df), dtype=np.bool_)
            leaf_pos_filter[internal_node_idxs] = False

            return phylogeny_df.loc[leaf_pos_filter, "id"].to_list()

    all_ids = ods.OrderedSet(phylogeny_df["id"])
    internal_ids = (
        set(
            ancestor_id
            for ancestor_list_str in phylogeny_df["ancestor_list"]
            for ancestor_id in alifestd_parse_ancestor_ids(ancestor_list_str)
        )
        if "ancestor_id" not in phylogeny_df
        else set(
            # root is self ref, but must exclude to handle only-root phylo
            phylogeny_df.loc[
                phylogeny_df["ancestor_id"] != phylogeny_df["id"],
                "ancestor_id",
            ]
        )
    )
    return list(all_ids - internal_ids)



"""Visualizations for stratum retention policies."""

from ._mrca_uncertainty_absolute_barplot import (
    mrca_uncertainty_absolute_barplot,
)
from ._mrca_uncertainty_relative_barplot import (
    mrca_uncertainty_relative_barplot,
)
from ._policy_panel_plot import policy_panel_plot
from ._strata_retained_frac_lineplot import strata_retained_frac_lineplot
from ._strata_retained_num_lineplot import strata_retained_num_lineplot
from ._stratum_retention_dripplot import stratum_retention_dripplot

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "mrca_uncertainty_absolute_barplot",
    "mrca_uncertainty_relative_barplot",
    "policy_panel_plot",
    "strata_retained_frac_lineplot",
    "strata_retained_num_lineplot",
    "stratum_retention_dripplot",
]

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import math

import anytree

from hstrat import hstrat
import hstrat.phylogenetic_inference.tree._impl as impl


def test_assign_trie_destruction_time_youngestplus1_single_leaf():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(),
        ],
    )(root, p_differentia_collision=0.5, mutate=True)
    assert leaf.destruction_time == math.inf
    assert inner.destruction_time == 1
    assert root.destruction_time == 1


def test_assign_trie_destruction_time_youngestplus1_two_leaves():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf1 = impl.TrieLeafNode(parent=inner, taxon_label="A")
    leaf2 = impl.TrieLeafNode(parent=inner, taxon_label="B")
    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(),
        ],
    )(root, p_differentia_collision=0.5, mutate=True)
    assert root.destruction_time == 1
    assert inner.destruction_time == 1
    assert leaf1.destruction_time == math.inf
    assert leaf2.destruction_time == math.inf


def test_assign_trie_destruction_time_youngestplus1_1():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=1, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(),
        ],
    )(root, p_differentia_collision=0.5, mutate=True)
    assert root.destruction_time == 1
    assert inner1.destruction_time == 2
    assert inner2.destruction_time == 2
    assert leaf.destruction_time == math.inf


def test_assign_trie_destruction_time_youngestplus1_2():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=2, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(),
        ],
    )(root, p_differentia_collision=0.5, mutate=True)
    assert root.destruction_time == 1.5
    assert inner1.destruction_time == 3
    assert inner2.destruction_time == 3
    assert leaf.destruction_time == math.inf


def test_assign_trie_destruction_time_youngestplus1_3():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    inner2 = impl.TrieInnerNode(rank=3, differentia=10, parent=inner1)
    leaf = impl.TrieLeafNode(parent=inner2, taxon_label="A")
    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(),
        ],
    )(root, p_differentia_collision=0.5, mutate=True)
    assert root.destruction_time == 2
    assert inner1.destruction_time == 4
    assert inner2.destruction_time == 4
    assert leaf.destruction_time == math.inf


def test_assign_trie_destruction_time_youngestplus1_complex():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner1 = impl.TrieInnerNode(rank=0, differentia=0, parent=root)

    inner2 = impl.TrieInnerNode(rank=4, differentia=10, parent=inner1)
    inner3 = impl.TrieInnerNode(rank=7, differentia=7, parent=inner2)

    inner2a = impl.TrieInnerNode(rank=1, differentia=1, parent=inner1)

    leaf3_a = impl.TrieLeafNode(parent=inner3, taxon_label="leaf3a_a")

    leaf2a_a = impl.TrieLeafNode(parent=inner2a, taxon_label="leaf2a_a")
    leaf2a_b = impl.TrieLeafNode(parent=inner2a, taxon_label="leaf2a_b")

    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(),
        ],
    )(root, p_differentia_collision=0.5, mutate=True)

    assert root.destruction_time == 1
    assert inner1.destruction_time == 2
    assert inner2.destruction_time == 8
    assert inner3.destruction_time == 8
    assert inner2a.destruction_time == 2
    assert leaf2a_a.destruction_time == math.inf
    assert leaf2a_b.destruction_time == math.inf
    assert leaf3_a.destruction_time == math.inf


def test_assign_trie_destruction_time_youngestplus1_assigned_property():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(
                assigned_property="blueberry"
            ),
        ],
    )(root, p_differentia_collision=0.5, mutate=True)
    assert leaf.blueberry == math.inf
    assert inner.blueberry == 1
    assert root.blueberry == 1


def test_assign_trie_destruction_time_youngestplus1_origin_time_property():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(
                assigned_property="blueberry"
            ),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(
                origin_time_property="blueberry"
            ),
        ]
    )(root, p_differentia_collision=0.5, mutate=True)
    assert leaf.destruction_time == math.inf
    assert inner.destruction_time == 1
    assert root.destruction_time == 1


def test_assign_trie_destruction_time_youngestplus1_otime_and_aprop():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(
                assigned_property="blueberry"
            ),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(
                origin_time_property="blueberry", assigned_property="raspberry"
            ),
        ],
    )(root, p_differentia_collision=0.5, mutate=True)
    assert leaf.raspberry == math.inf
    assert inner.raspberry == 1
    assert root.raspberry == 1


def test_assign_trie_destruction_time_youngestplus1_mutate():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root_ = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(
                assigned_property="blueberry"
            ),
        ],
    )(root, p_differentia_collision=0.5, mutate=False)
    assert not hasattr(leaf, "blueberry")
    assert not hasattr(inner, "blueberry")
    assert not hasattr(root, "blueberry")

    assert hasattr(root_, "blueberry")

    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNaiveTriePostprocessor(),
            hstrat.AssignDestructionTimeYoungestPlusOneTriePostprocessor(
                assigned_property="blueberry"
            ),
        ],
    )(root, p_differentia_collision=0.5, mutate=True)
    assert str(anytree.RenderTree(root).by_attr("blueberry")) == str(
        anytree.RenderTree(root_).by_attr("blueberry")
    )



import unittest

from hstrat._auxiliary_lib import splicewhile


class TestConsume(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_iterable(self):
        for condition in True, False:
            try:
                splicewhile(lambda x: condition, range(100))
            except TypeError:
                pass
            else:
                assert False

            try:
                splicewhile(lambda x: condition, [1, 2, 3, 4])
            except TypeError:
                pass
            else:
                assert False

            try:
                splicewhile(lambda x: condition, [])
            except TypeError:
                pass
            else:
                assert False

    def test_empty(self):
        for condition in True, False:
            it = iter([])
            res, it = splicewhile(lambda x: condition, it)
            assert res == []
            for __ in it:
                assert False

        for condition in True, False:
            it = iter(range(0))
            res, it = splicewhile(lambda x: condition, it)
            for __ in it:
                assert False

    def test_singleton(self):
        it = iter("a")
        res, it = splicewhile(lambda x: x == "a", it)
        assert res == ["a"]
        assert [*it] == []

        it = iter("a")
        res, it = splicewhile(lambda x: x != "a", it)
        assert res == []
        assert [*it] == ["a"]

        for n in 1, 2, 3:
            it = iter(range(1))
            res, it = splicewhile(lambda x: x < n, it)
            assert res == [0]
            assert [*it] == []

    def test_conservation(self):
        for n in range(20):
            for cut in range(n + 1):
                it = iter(range(n))
                res, it = splicewhile(lambda x: x != cut, it)
                rest = [*it]
                assert res == [*range(cut)]
                assert res + rest == [*range(n)]


if __name__ == "__main__":
    unittest.main()



import typing

from .._detail import PolicySpecBase


class PolicySpec(PolicySpecBase):
    """Contains all policy parameters, if any."""

    _depth_proportional_resolution: int

    def __init__(
        self: "PolicySpec",
        depth_proportional_resolution: int,
    ):
        """Construct the policy spec.

        Parameters
        ----------
        depth_proportional_resolution : int
            The desired minimum number of intervals for the rank of the MRCA to
            be able to be distinguished between. The uncertainty of MRCA
            rank estimates provided under the depth-proportional resolution
            policy will scale as total number of strata deposited divided by
            depth_proportional_resolution.
        """
        assert depth_proportional_resolution > 0
        self._depth_proportional_resolution = depth_proportional_resolution

    def __eq__(self: "PolicySpec", other: typing.Any) -> bool:
        return isinstance(other, self.__class__) and (
            self._depth_proportional_resolution,
        ) == (other._depth_proportional_resolution,)

    def __repr__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoIdentifier()
        }.{
            PolicySpec.__qualname__
        }(depth_proportional_resolution={
            self._depth_proportional_resolution
        })"""

    def __str__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoTitle()
        } (resolution: {
            self._depth_proportional_resolution
        })"""

    def GetEvalCtor(self: "PolicySpec") -> str:
        return f"hstrat.{self!r}"

    def GetDepthProportionalResolution(self: "PolicySpec") -> int:
        return self._depth_proportional_resolution

    @staticmethod
    def GetAlgoIdentifier() -> str:
        """Get programatic name for underlying retention algorithm."""
        return __package__.split(".")[-1]

    @staticmethod
    def GetAlgoTitle() -> str:
        """Get human-readable name for underlying retention algorithm."""
        return "Depth-proportional Resolution Stratum Retention Algorithm"



from hstrat._auxiliary_lib import unzip


def test_unzip_with_two_element_tuples():
    input_iterable = [(1, 2), (3, 4), (5, 6)]
    expected_output = [(1, 3, 5), (2, 4, 6)]
    assert list(unzip(input_iterable)) == expected_output


def test_unzip_with_three_element_tuples():
    input_iterable = [(1, 2, 3), (4, 5, 6)]
    expected_output = [(1, 4), (2, 5), (3, 6)]
    assert list(unzip(input_iterable)) == expected_output


def test_unzip_with_one_element_tuples():
    input_iterable = [(1,), (2,), (3,)]
    expected_output = [
        (1, 2, 3),
    ]
    assert list(unzip(input_iterable)) == expected_output


def test_unzip_with_empty_tuples():
    input_iterable = [(), (), ()]
    expected_output = []
    assert list(unzip(input_iterable)) == expected_output


def test_unzip_with_empty():
    input_iterable = []
    expected_output = []
    assert list(unzip(input_iterable)) == expected_output



import functools
import json
import random

import pytest
from tqdm import tqdm

from hstrat import genome_instrumentation, hstrat


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
@pytest.mark.parametrize(
    "popsize",
    [0, 1, 8],
)
def test_pop_to_records(
    impl,
    retention_policy,
    ordered_store,
    differentia_bit_width,
    popsize,
):

    pop = [
        impl(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
            stratum_differentia_bit_width=differentia_bit_width,
        )
        for __ in range(popsize)
    ]
    for _ in range(100):
        random.shuffle(pop)
        if pop:
            pop[0] = pop[-1].CloneDescendant()

    assert hstrat.pop_to_records(pop) == hstrat.pop_to_records(pop)
    records = hstrat.pop_to_records(
        pop, progress_wrap=functools.partial(tqdm, disable=True)
    )
    for field in (
        "policy",
        "policy_algo",
        "policy_spec",
        "differentia_bit_width",
        "hstrat_version",
    ):
        assert field in records
        assert not any(field in col_rec for col_rec in records["columns"])

    reconstituted = hstrat.pop_from_records(
        records, progress_wrap=functools.partial(tqdm, disable=True)
    )
    if (
        ordered_store == hstrat.HereditaryStratumOrderedStoreList
        and impl == genome_instrumentation.HereditaryStratigraphicColumn
    ):
        assert reconstituted == pop
    else:
        assert hstrat.pop_to_records(reconstituted) == hstrat.pop_to_records(
            pop
        )

    records = hstrat.pop_to_records(pop)
    json_str = json.dumps(records)
    reconstituted = hstrat.pop_from_records(
        json.loads(json_str),
        progress_wrap=functools.partial(tqdm, disable=True),
    )
    if (
        ordered_store == hstrat.HereditaryStratumOrderedStoreList
        and impl == genome_instrumentation.HereditaryStratigraphicColumn
    ):
        assert reconstituted == pop
    else:
        assert hstrat.pop_to_records(reconstituted) == hstrat.pop_to_records(
            pop
        )


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
@pytest.mark.parametrize(
    "popsize",
    [0, 1, 8],
)
def test_pop_to_records_then_from_records(
    impl,
    retention_policy,
    ordered_store,
    differentia_bit_width,
    popsize,
):

    pop = [
        impl(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
            stratum_differentia_bit_width=differentia_bit_width,
        )
        for __ in range(popsize)
    ]
    for _ in range(100):
        random.shuffle(pop)
        if pop:
            pop[0] = pop[-1].CloneDescendant()

    records = hstrat.pop_to_records(pop)
    reconstituted = hstrat.pop_from_records(records)
    if (
        ordered_store == hstrat.HereditaryStratumOrderedStoreList
        and impl == genome_instrumentation.HereditaryStratigraphicColumn
    ):
        assert reconstituted == pop
    else:
        assert hstrat.pop_to_records(reconstituted) == hstrat.pop_to_records(
            pop, progress_wrap=functools.partial(tqdm, disable=True)
        )


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
@pytest.mark.parametrize(
    "popsize",
    [0, 1, 8],
)
def test_pop_to_records_then_from_records_json(
    impl,
    retention_policy,
    ordered_store,
    differentia_bit_width,
    popsize,
):

    pop = [
        impl(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
            stratum_differentia_bit_width=differentia_bit_width,
        )
        for __ in range(popsize)
    ]
    for _ in range(100):
        random.shuffle(pop)
        if pop:
            pop[0] = pop[-1].CloneDescendant()

    records = hstrat.pop_to_records(pop)
    json_str = json.dumps(records)
    reconstituted = hstrat.pop_from_records(json.loads(json_str))
    if (
        ordered_store == hstrat.HereditaryStratumOrderedStoreList
        and impl == genome_instrumentation.HereditaryStratigraphicColumn
    ):
        assert reconstituted == pop
    else:
        assert hstrat.pop_to_records(reconstituted) == hstrat.pop_to_records(
            pop
        )



from collections import Counter, defaultdict
import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_assign_contiguous_ids,
    alifestd_collapse_unifurcations,
    alifestd_find_leaf_ids,
    alifestd_find_root_ids,
    alifestd_is_asexual,
    alifestd_is_sexual,
    alifestd_is_topologically_sorted,
    alifestd_parse_ancestor_ids,
    alifestd_to_working_format,
    alifestd_topological_sort,
    alifestd_try_add_ancestor_id_col,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
                ),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_assign_contiguous_ids,
        alifestd_to_working_format,
        alifestd_topological_sort,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
@pytest.mark.parametrize(
    "root_ancestor_token",
    [
        "",
        "None",
        "none",
    ],
)
def test_alifestd_collapse_unifurcations(
    phylogeny_df, apply, root_ancestor_token
):
    phylogeny_df = apply(phylogeny_df.copy())

    phylogeny_df_ = phylogeny_df.copy()
    collapsed_df = alifestd_collapse_unifurcations(
        phylogeny_df, root_ancestor_token=root_ancestor_token
    )
    assert len(alifestd_find_root_ids(phylogeny_df)) == sum(
        collapsed_df["ancestor_list"] == f"[{root_ancestor_token}]"
    )
    assert alifestd_validate(collapsed_df)
    assert phylogeny_df.equals(phylogeny_df_)

    assert alifestd_is_asexual(collapsed_df) == alifestd_is_asexual(
        phylogeny_df
    )
    assert alifestd_is_sexual(collapsed_df) == alifestd_is_sexual(phylogeny_df)
    if alifestd_is_asexual(phylogeny_df):
        assert len(collapsed_df) < len(phylogeny_df)

    assert set(alifestd_find_leaf_ids(phylogeny_df)) == set(
        alifestd_find_leaf_ids(collapsed_df)
    )

    phylogeny_df_ = phylogeny_df.set_index("id", drop=False)
    for _idx, row in collapsed_df.iterrows():
        assert all(
            phylogeny_df_.loc[row["id"]]
            .drop(["ancestor_list", "ancestor_id"], errors="ignore")
            .dropna()
            == row.drop(
                ["ancestor_list", "ancestor_id"], errors="ignore"
            ).dropna()
        )

    ref_counts = Counter(
        id_
        for ancestor_list_str in collapsed_df["ancestor_list"]
        for id_ in alifestd_parse_ancestor_ids(ancestor_list_str)
    )
    for _idx, row in collapsed_df.iterrows():
        if len(alifestd_parse_ancestor_ids(row["ancestor_list"])) == 1:
            assert ref_counts[row["id"]] != 1

    phylogeny_df = alifestd_topological_sort(phylogeny_df)
    assert alifestd_is_topologically_sorted(collapsed_df)

    collapsed_descendants_lookup = defaultdict(set)
    phylogeny_descendants_lookup = defaultdict(set)

    for df, lookup in (
        (phylogeny_df, phylogeny_descendants_lookup),
        (collapsed_df, collapsed_descendants_lookup),
    ):
        for _idx, row in df[::-1].iterrows():
            id_ = row["id"]
            ancestor_ids = alifestd_parse_ancestor_ids(row["ancestor_list"])
            lookup[id_].add(id_)
            for ancestor_id in ancestor_ids:
                lookup[ancestor_id] |= lookup[id_]

    dropped_ids = set(phylogeny_df["id"]) - set(collapsed_df["id"])
    for id_, descendants in collapsed_descendants_lookup.items():
        assert descendants <= phylogeny_descendants_lookup[id_]
        assert descendants == (phylogeny_descendants_lookup[id_] - dropped_ids)
